"""Tests for pyotgw/connection.py"""

import asyncio
import functools
import logging
from unittest.mock import DEFAULT, MagicMock, patch

import pytest
import serial

from pyotgw.connection import MAX_RETRY_TIMEOUT
from pyotgw.protocol import OpenThermProtocol
from tests.helpers import called_once, called_x_times


@pytest.mark.asyncio
async def test_connect_success_and_reconnect(caplog, pygw_conn, pygw_proto):
    """Test ConnectionManager.connect()"""
    pygw_conn._error = asyncio.CancelledError()

    with patch.object(
        pygw_conn,
        "_attempt_connect",
        return_value=(pygw_proto.transport, pygw_proto),
    ) as attempt_connect, caplog.at_level(logging.DEBUG):
        assert await pygw_conn.connect("loop://")

    assert pygw_conn._port == "loop://"
    attempt_connect.assert_called_once()
    assert pygw_conn._connecting_task is None
    assert pygw_conn._error is None
    assert caplog.record_tuples == [
        (
            "pyotgw.connection",
            logging.DEBUG,
            "Connected to serial device on loop://",
        ),
    ]
    assert pygw_conn._transport == pygw_proto.transport
    assert pygw_conn.protocol == pygw_proto
    assert pygw_conn.watchdog.is_active
    assert pygw_conn.connected

    await pygw_conn.watchdog.stop()
    caplog.clear()

    with patch.object(
        pygw_conn,
        "_attempt_connect",
        return_value=(pygw_proto.transport, pygw_proto),
    ) as attempt_connect, caplog.at_level(logging.DEBUG):
        assert await pygw_conn.connect("loop://new")

    assert pygw_conn._port == "loop://new"
    attempt_connect.assert_called_once()
    assert pygw_conn._connecting_task is None
    assert pygw_conn._error is None
    assert caplog.record_tuples == [
        (
            "pyotgw.connection",
            logging.DEBUG,
            "Reconnecting to serial device on loop://new",
        ),
        (
            "pyotgw.connection",
            logging.DEBUG,
            "Connected to serial device on loop://new",
        ),
    ]
    assert pygw_conn._transport == pygw_proto.transport
    assert pygw_conn.protocol == pygw_proto
    assert pygw_conn.watchdog.is_active
    assert pygw_conn.connected


@pytest.mark.asyncio
async def test_connect_cancel(pygw_conn):
    """Test ConnectionManager.connect() cancellation"""
    with patch.object(
        pygw_conn,
        "_attempt_connect",
        side_effect=asyncio.CancelledError,
    ) as create_serial_connection:
        assert not await pygw_conn.connect("loop://")

    create_serial_connection.assert_called_once()


@pytest.mark.asyncio
async def test_disconnect(pygw_conn, pygw_proto):
    """Test ConnectionManager.disconnect()"""
    with patch.object(
        pygw_conn,
        "_attempt_connect",
        return_value=(pygw_proto.transport, pygw_proto),
    ):
        assert await pygw_conn.connect("loop://")

    with patch.object(
        pygw_proto,
        "disconnect",
    ) as disconnect:
        await pygw_conn.disconnect()
    assert disconnect.called


@pytest.mark.asyncio
async def test_disconnect_while_connecting(pygw_conn):
    """Test ConnectionManager.disconnect() during an ongoing connection attempt"""
    loop = asyncio.get_running_loop()

    async def wait_forever():
        while True:
            await asyncio.sleep(1)

    with patch.object(
        pygw_conn,
        "_attempt_connect",
        side_effect=wait_forever,
    ) as attempt_connect:
        task = loop.create_task(pygw_conn.connect("loop://"))
        await called_once(attempt_connect)
        await pygw_conn.disconnect()

        assert await task is False
        assert not pygw_conn.connected


@pytest.mark.asyncio
async def test_reconnect(caplog, pygw_conn):
    """Test ConnectionManager.reconnect()"""
    with patch.object(pygw_conn, "disconnect") as disconnect, patch.object(
        pygw_conn,
        "connect",
    ) as connect, caplog.at_level(logging.ERROR):
        await pygw_conn.reconnect()

    disconnect.assert_not_called()
    connect.assert_not_called()
    assert caplog.record_tuples == [
        ("pyotgw.connection", logging.ERROR, "Reconnect called before connect!")
    ]

    caplog.clear()
    pygw_conn._port = "loop://"
    with patch.object(pygw_conn, "disconnect") as disconnect, patch.object(
        pygw_conn._otgw,
        "connect",
    ) as connect, caplog.at_level(logging.DEBUG):
        await pygw_conn.reconnect()

    assert caplog.record_tuples == [
        ("pyotgw.connection", logging.DEBUG, "Scheduling reconnect...")
    ]
    disconnect.assert_called_once()
    connect.assert_called_once_with("loop://")


@pytest.mark.asyncio
async def test_reconnect_after_connection_loss(caplog, pygw_conn, pygw_proto):
    """Test ConnectionManager.reconnect() after connection loss"""
    pygw_conn._error = asyncio.CancelledError()

    with patch.object(
        pygw_conn._otgw,
        "connect",
        side_effect=pygw_conn.connect,
    ), patch.object(
        pygw_conn,
        "_attempt_connect",
        return_value=(pygw_proto.transport, pygw_proto),
    ) as attempt_conn, patch.object(
        pygw_conn.watchdog,
        "start",
        side_effect=pygw_conn.watchdog.start,
    ) as wd_start, caplog.at_level(
        logging.DEBUG
    ):
        assert await pygw_conn.connect("loop://", timeout=0.001)

        caplog.clear()
        attempt_conn.assert_called_once()
        attempt_conn.reset_mock()
        await called_x_times(wd_start, 2, timeout=3)
        assert pygw_conn.protocol.connected
        attempt_conn.assert_called_once()
        assert caplog.record_tuples == [
            (
                "pyotgw.connection",
                logging.DEBUG,
                "Watchdog triggered!",
            ),
            (
                "pyotgw.connection",
                logging.DEBUG,
                "Canceling Watchdog task.",
            ),
            (
                "pyotgw.connection",
                logging.DEBUG,
                "Scheduling reconnect...",
            ),
            (
                "pyotgw.connection",
                logging.DEBUG,
                "Reconnecting to serial device on loop://",
            ),
            (
                "pyotgw.connection",
                logging.DEBUG,
                "Connected to serial device on loop://",
            ),
        ]


def test_connected(pygw_conn, pygw_proto):
    """Test ConnectionManager.connected()"""
    pygw_conn.protocol = pygw_proto
    pygw_conn.protocol._connected = False
    assert not pygw_conn.connected
    pygw_conn.protocol._connected = True
    assert pygw_conn.connected


def test_set_connection_config(pygw_conn):
    """Test ConnectionManager.set_connection_config()"""
    assert pygw_conn.set_connection_config(baudrate=19200, parity=serial.PARITY_NONE)
    assert pygw_conn._config == {
        "baudrate": 19200,
        "bytesize": serial.EIGHTBITS,
        "parity": serial.PARITY_NONE,
        "stopbits": serial.STOPBITS_ONE,
    }

    assert not pygw_conn.set_connection_config(baudrate=9600, invalid="value")
    assert pygw_conn._config == {
        "baudrate": 19200,
        "bytesize": serial.EIGHTBITS,
        "parity": serial.PARITY_NONE,
        "stopbits": serial.STOPBITS_ONE,
    }

    with patch("pyotgw.connection.ConnectionManager.connected", return_value=True):
        assert not pygw_conn.set_connection_config()


@pytest.mark.asyncio
async def test_attempt_connect_success(pygw_conn, pygw_proto):
    """Test ConnectionManager._attempt_connect()"""
    pygw_conn._port = "loop://"

    # We cannot compare the functools.partial object which is created in-line
    # so we have to manually save args and kwargs and compare them one by one.
    saved_args_list = []

    def save_args(*used_args, **used_kwargs):
        """Store args and kwargs each time we're called"""
        saved_args_list.append({"args": used_args, "kwargs": used_kwargs})
        return DEFAULT

    with patch(
        "pyotgw.protocol.OpenThermProtocol.init_and_wait_for_activity",
    ) as init_and_wait, patch(
        "serial_asyncio_fast.create_serial_connection",
        return_value=(pygw_proto.transport, pygw_proto),
        side_effect=save_args,
    ) as create_serial_connection:
        assert await pygw_conn._attempt_connect() == (pygw_proto.transport, pygw_proto)

    create_serial_connection.assert_called_once()
    assert len(saved_args_list) == 1
    args = saved_args_list[0]["args"]
    assert len(args) == 3
    assert args[0] == asyncio.get_running_loop()
    assert isinstance(args[1], functools.partial)
    assert args[1].func == OpenThermProtocol
    assert args[1].args == (
        pygw_conn._otgw.status,
        pygw_conn.watchdog.inform,
    )
    assert args[2] == pygw_conn._port
    kwargs = saved_args_list[0]["kwargs"]
    assert kwargs == {
        "write_timeout": 0,
        "baudrate": 9600,
        "bytesize": serial.EIGHTBITS,
        "parity": serial.PARITY_NONE,
        "stopbits": serial.STOPBITS_ONE,
    }
    init_and_wait.assert_called_once()


@pytest.mark.asyncio
async def test_attempt_connect_oserror(caplog, pygw_conn):
    """Test ConnectionManager._attempt_connect() with OSError"""
    loop = asyncio.get_running_loop()
    pygw_conn._port = "loop://"

    with patch(
        "serial_asyncio_fast.create_serial_connection",
        side_effect=OSError,
    ) as create_serial_connection, patch.object(
        pygw_conn,
        "_get_retry_timeout",
        return_value=0,
    ) as retry_timeout, caplog.at_level(
        logging.ERROR
    ):
        task = loop.create_task(pygw_conn._attempt_connect())
        await called_x_times(retry_timeout, 2)

        assert create_serial_connection.call_count >= 2
        assert caplog.record_tuples == [
            (
                "pyotgw.connection",
                logging.ERROR,
                "Could not connect to serial device on loop://. "
                "Will keep trying. Reported error was: ",
            )
        ]

        task.cancel()
        with pytest.raises(asyncio.CancelledError):
            await task


@pytest.mark.asyncio
async def test_attempt_connect_serialexception(caplog, pygw_conn):
    """Test ConnectionManager._attempt_connect() with SerialException"""
    loop = asyncio.get_running_loop()
    pygw_conn._port = "loop://"

    with patch(
        "serial_asyncio_fast.create_serial_connection",
        side_effect=serial.SerialException,
    ) as create_serial_connection, patch.object(
        pygw_conn,
        "_get_retry_timeout",
        return_value=0,
    ) as retry_timeout, caplog.at_level(
        logging.ERROR
    ):
        task = loop.create_task(pygw_conn._attempt_connect())
        await called_x_times(retry_timeout, 2)

        assert create_serial_connection.call_count >= 2
        assert caplog.record_tuples == [
            (
                "pyotgw.connection",
                logging.ERROR,
                "Could not connect to serial device on loop://. "
                "Will keep trying. Reported error was: ",
            )
        ]

        task.cancel()
        with pytest.raises(asyncio.CancelledError):
            await task


@pytest.mark.asyncio
async def test_attempt_connect_timeouterror(caplog, pygw_conn, pygw_proto):
    """Test ConnectionManager._attempt_connect() with SerialException"""
    loop = asyncio.get_running_loop()
    pygw_conn._port = "loop://"

    pygw_proto.init_and_wait_for_activity = MagicMock(side_effect=asyncio.TimeoutError)
    pygw_proto.disconnect = MagicMock()

    with patch(
        "serial_asyncio_fast.create_serial_connection",
        return_value=(pygw_proto.transport, pygw_proto),
    ) as create_serial_connection, patch.object(
        pygw_conn,
        "_get_retry_timeout",
        return_value=0,
    ) as retry_timeout, caplog.at_level(
        logging.ERROR
    ):
        task = loop.create_task(pygw_conn._attempt_connect())
        await called_x_times(retry_timeout, 2)

        assert create_serial_connection.call_count >= 2
        assert pygw_proto.disconnect.call_count >= 2
        assert caplog.record_tuples == [
            (
                "pyotgw.connection",
                logging.ERROR,
                "The serial device on loop:// is not responding. " "Will keep trying.",
            )
        ]

        task.cancel()
        with pytest.raises(asyncio.CancelledError):
            await task


@pytest.mark.asyncio
async def test_attempt_connect_syntaxerror(caplog, pygw_conn, pygw_proto):
    """Test ConnectionManager._attempt_connect() with SyntaxError"""
    loop = asyncio.get_running_loop()
    pygw_conn._port = "loop://"

    pygw_proto.init_and_wait_for_activity = MagicMock(side_effect=SyntaxError)
    pygw_proto.disconnect = MagicMock()

    with patch(
        "serial_asyncio_fast.create_serial_connection",
        return_value=(pygw_proto.transport, pygw_proto),
    ) as create_serial_connection, patch.object(
        pygw_conn,
        "_get_retry_timeout",
        return_value=0,
    ) as retry_timeout, caplog.at_level(
        logging.ERROR
    ):
        task = loop.create_task(pygw_conn._attempt_connect())
        with pytest.raises(SyntaxError):
            await task

        assert retry_timeout.call_count == 1
        assert create_serial_connection.call_count >= 2
        assert pygw_proto.disconnect.call_count >= 2
        assert isinstance(pygw_conn._error, SyntaxError)


@pytest.mark.asyncio
async def test_cleanup(pygw_conn):
    """Test ConnectionManager._cleanup()"""
    pass  # with patch.object()


def test_get_retry_timeout(pygw):
    """Test pyotgw._get_retry_timeout()"""
    pygw.connection._retry_timeout = MAX_RETRY_TIMEOUT / 2
    assert pygw.connection._get_retry_timeout() == MAX_RETRY_TIMEOUT / 2
    assert pygw.connection._get_retry_timeout() == (MAX_RETRY_TIMEOUT / 2) * 1.5
    assert pygw.connection._get_retry_timeout() == MAX_RETRY_TIMEOUT


def test_is_active(pygw_watchdog):
    """Test ConnectionWatchdog.is_active()"""
    assert not pygw_watchdog.is_active
    pygw_watchdog.start(None, 10)
    assert pygw_watchdog.is_active


@pytest.mark.asyncio
async def test_inform_watchdog(caplog, pygw_watchdog):
    """Test ConnectionWatchdog.inform()"""
    await pygw_watchdog.inform()

    async def empty_coroutine():
        return

    pygw_watchdog.start(empty_coroutine, 10)

    with patch.object(
        pygw_watchdog._wd_task,
        "cancel",
        side_effect=pygw_watchdog._wd_task.cancel,
    ) as task_cancel, patch.object(
        pygw_watchdog,
        "_watchdog",
    ) as watchdog, caplog.at_level(
        logging.DEBUG
    ):
        await pygw_watchdog.inform()

    task_cancel.assert_called_once()
    watchdog.assert_called_once_with(pygw_watchdog.timeout)
    assert caplog.record_tuples == [
        ("pyotgw.connection", logging.DEBUG, "Watchdog timer reset!")
    ]


def test_start(pygw_watchdog):
    """Test ConnectionWatchdog.start()"""

    def callback():
        return

    with patch.object(
        pygw_watchdog,
        "_watchdog",
    ) as watchdog:
        assert pygw_watchdog.start(callback, 10)

    assert pygw_watchdog.timeout == 10
    assert pygw_watchdog._callback == callback
    watchdog.assert_called_once_with(10)

    assert not pygw_watchdog.start(callback, 10)


@pytest.mark.asyncio
async def test_stop(caplog, pygw_watchdog):
    """Test ConnectionWatchdog.stop()"""
    with caplog.at_level(logging.DEBUG):
        await pygw_watchdog.stop()

    assert caplog.records == []

    async def empty_coroutine():
        return

    pygw_watchdog.start(empty_coroutine, 10)
    with caplog.at_level(logging.DEBUG):
        await pygw_watchdog.stop()

    assert not pygw_watchdog.is_active
    assert caplog.record_tuples == [
        (
            "pyotgw.connection",
            logging.DEBUG,
            "Canceling Watchdog task.",
        ),
    ]


@pytest.mark.asyncio
async def test_watchdog(caplog, pygw_watchdog):
    """Test ConnectionWatchdog._watchdog()"""

    async def empty_callback():
        return

    watchdog_callback = MagicMock(side_effect=empty_callback)
    pygw_watchdog.start(watchdog_callback, 0)

    with caplog.at_level(logging.DEBUG):
        await called_once(watchdog_callback)

    assert caplog.record_tuples == [
        (
            "pyotgw.connection",
            logging.DEBUG,
            "Watchdog triggered!",
        ),
        (
            "pyotgw.connection",
            logging.DEBUG,
            "Canceling Watchdog task.",
        ),
    ]



"""pyotgw is a library to interface with the OpenTherm Gateway."""

import asyncio
import logging
from datetime import datetime

from pyotgw import vars as v
from pyotgw.connection import ConnectionManager
from pyotgw.status import StatusManager

_LOGGER = logging.getLogger(__name__)


class OpenThermGateway:  # pylint: disable=too-many-public-methods
    """Main OpenThermGateway object abstraction"""

    def __init__(self):
        """Create an OpenThermGateway object."""
        self._transport = None
        self._protocol = None
        self._gpio_task = None
        self._skip_init = False
        self.status = StatusManager()
        self.connection = ConnectionManager(self)

    async def cleanup(self):
        """Clean up tasks."""
        await self.connection.disconnect()
        await self.status.cleanup()
        if self._gpio_task:
            self._gpio_task.cancel()
            await self._gpio_task

    async def connect(
        self,
        port,
        timeout=None,
        skip_init=None,
    ):
        """
        Connect to Opentherm Gateway at @port.
        Initialize the parameters obtained from the PS= and PR=
        commands and returns the status dict with the obtained values
        or False if cancelled.
        If @skip_init is True, the PS= and PR= commands are skipped and only PS=0 is
        sent upon this and future connection attempts.
        If called while connected, reconnect to the gateway.

        This method is a coroutine
        """
        if skip_init is not None:
            self._skip_init = skip_init
        if not await self.connection.connect(port, timeout):
            return False
        self._protocol = self.connection.protocol
        if not self._skip_init:
            await self.get_reports()
            await self.get_status()
        await self._poll_gpio()
        return self.status.status

    async def disconnect(self):
        """Disconnect from the OpenTherm Gateway."""
        await self.cleanup()
        return await self.connection.disconnect()

    def set_connection_options(self, **kwargs):
        """Set connection parameters."""
        return self.connection.set_connection_config(**kwargs)

    async def set_target_temp(
        self, temp, temporary=True, timeout=v.OTGW_DEFAULT_TIMEOUT
    ):
        """
        Set the thermostat setpoint and return the newly accepted
        value.
        kwarg @temporary specifies whether or not the thermostat
            program may override this temperature.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_TARGET_TEMP if temporary else v.OTGW_CMD_TARGET_TEMP_CONST
        value = f"{temp:2.1f}"
        ret = await self._wait_for_cmd(cmd, value, timeout)
        if ret is None:
            return None
        ret = float(ret)
        if 0 <= ret <= 30:
            if ret == 0:
                status_update = {
                    v.OTGW: {v.OTGW_SETP_OVRD_MODE: v.OTGW_SETP_OVRD_DISABLED},
                    v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: None},
                }
            else:
                if temporary:
                    ovrd_mode = v.OTGW_SETP_OVRD_TEMPORARY
                else:
                    ovrd_mode = v.OTGW_SETP_OVRD_PERMANENT
                status_update = {
                    v.OTGW: {v.OTGW_SETP_OVRD_MODE: ovrd_mode},
                    v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: ret},
                }
            self.status.submit_full_update(status_update)
            return ret

    async def set_temp_sensor_function(self, func, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Set the temperature sensor function. The following functions are available:
            O: Outside temperature
            R: Return water temperature

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_TEMP_SENSOR
        if func not in "OR":
            return None
        ret = await self._wait_for_cmd(cmd, func, timeout)
        if ret is None:
            return None
        status_otgw = {v.OTGW_TEMP_SENSOR: ret}
        self.status.submit_partial_update(v.OTGW, status_otgw)
        return ret

    async def set_outside_temp(self, temp, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Configure the outside temperature to send to the thermostat.
        Allowed values are between -40.0 and +64.0, although
        thermostats may not display the full range. Specify a value
        above 64 (suggestion: 99) to clear a previously configured
        value.
        Return the accepted value on success, '-' if a previously
        configured value has been cleared or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_OUTSIDE_TEMP
        status_thermostat = {}
        if temp < -40:
            return None
        value = f"{temp:2.1f}"
        ret = await self._wait_for_cmd(cmd, value, timeout)
        if ret is None:
            return
        if ret == "-":
            status_thermostat[v.DATA_OUTSIDE_TEMP] = 0.0
        else:
            ret = float(ret)
            status_thermostat[v.DATA_OUTSIDE_TEMP] = ret
        self.status.submit_partial_update(v.THERMOSTAT, status_thermostat)
        return ret

    async def set_clock(self, date=datetime.now(), timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Change the time and day of the week of the thermostat. The
        gateway will send the specified time and day of the week in
        response to the next time and date message from the thermostat.
        @date is a :datetime: object which defaults to now()
        Return the response from the gateway with format HH:MM/DOW,
        where DOW is a single digit: 1=Monday, 7=Sunday.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_SET_CLOCK
        value = f"{date.strftime('%H:%M')}/{date.isoweekday()}"
        return await self._wait_for_cmd(cmd, value, timeout)

    async def get_reports(self):
        """
        Update the OpenThermGateway object with the information from all
        of the PR commands and return the updated status dict.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_REPORT
        reports = {}
        # Get version info first
        ret = await self._wait_for_cmd(cmd, v.OTGW_REPORT_ABOUT)
        reports[v.OTGW_REPORT_ABOUT] = ret[2:] if ret else None
        for value in v.OTGW_REPORTS:

            ver = reports.get(v.OTGW_REPORT_ABOUT)
            if ver and int(ver[18]) < 5 and value == "D":
                # Added in v5
                continue
            if value == v.OTGW_REPORT_ABOUT:
                continue

            ret = await self._wait_for_cmd(cmd, value)
            if ret is None:
                reports[value] = None
                continue
            reports[value] = ret[2:]

        status_otgw = {
            v.OTGW_ABOUT: reports.get(v.OTGW_REPORT_ABOUT),
            v.OTGW_BUILD: reports.get(v.OTGW_REPORT_BUILDDATE),
            v.OTGW_CLOCKMHZ: reports.get(v.OTGW_REPORT_CLOCKMHZ),
            v.OTGW_DHW_OVRD: reports.get(v.OTGW_REPORT_DHW_SETTING),
            v.OTGW_MODE: reports.get(v.OTGW_REPORT_GW_MODE),
            v.OTGW_RST_CAUSE: reports.get(v.OTGW_REPORT_RST_CAUSE),
            v.OTGW_SMART_PWR: reports.get(v.OTGW_REPORT_SMART_PWR),
            v.OTGW_TEMP_SENSOR: reports.get(v.OTGW_REPORT_TEMP_SENSOR),
            v.OTGW_THRM_DETECT: reports.get(v.OTGW_REPORT_THERMOSTAT_DETECT),
        }
        status_thermostat = {}
        ovrd_mode = reports.get(v.OTGW_REPORT_SETPOINT_OVRD)
        if ovrd_mode is not None:
            ovrd_mode = str.upper(ovrd_mode[0])
            status_otgw.update({v.OTGW_SETP_OVRD_MODE: ovrd_mode})
        gpio_funcs = reports.get(v.OTGW_REPORT_GPIO_FUNCS)
        if gpio_funcs is not None:
            status_otgw.update(
                {v.OTGW_GPIO_A: int(gpio_funcs[0]), v.OTGW_GPIO_B: int(gpio_funcs[1])}
            )
        led_funcs = reports.get(v.OTGW_REPORT_LED_FUNCS)
        if led_funcs is not None:
            status_otgw.update(
                {
                    v.OTGW_LED_A: led_funcs[0],
                    v.OTGW_LED_B: led_funcs[1],
                    v.OTGW_LED_C: led_funcs[2],
                    v.OTGW_LED_D: led_funcs[3],
                    v.OTGW_LED_E: led_funcs[4],
                    v.OTGW_LED_F: led_funcs[5],
                }
            )
        tweaks = reports.get(v.OTGW_REPORT_TWEAKS)
        if tweaks is not None:
            status_otgw.update(
                {
                    v.OTGW_IGNORE_TRANSITIONS: int(tweaks[0]),
                    v.OTGW_OVRD_HB: int(tweaks[1]),
                }
            )
        sb_temp = reports.get(v.OTGW_REPORT_SETBACK_TEMP)
        if sb_temp is not None:
            status_otgw.update({v.OTGW_SB_TEMP: float(sb_temp)})
        vref = reports.get(v.OTGW_REPORT_VREF)
        if vref is not None:
            status_otgw.update({v.OTGW_VREF: int(vref)})
        if ovrd_mode is not None and ovrd_mode != v.OTGW_SETP_OVRD_DISABLED:
            status_thermostat = {
                v.DATA_ROOM_SETPOINT_OVRD: float(
                    reports[v.OTGW_REPORT_SETPOINT_OVRD][1:]
                )
            }
        self.status.submit_full_update(
            {v.THERMOSTAT: status_thermostat, v.OTGW: status_otgw}
        )
        return self.status.status

    async def get_status(self):
        """
        Update the OpenThermGateway object with the information from the
        PS command and return the updated status dict.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_SUMMARY
        ret = await self._wait_for_cmd(cmd, 1)
        # Return to 'reporting' mode
        if ret is None:
            return
        asyncio.get_running_loop().create_task(self._wait_for_cmd(cmd, 0))
        fields = ret[1].split(",")
        if len(fields) == 34:
            # OpenTherm Gateway 5.0
            boiler_status, thermostat_status = process_statusfields_v5(fields)
        else:
            boiler_status, thermostat_status = process_statusfields_v4(fields)
        self.status.submit_full_update({
            v.BOILER: boiler_status,
            v.THERMOSTAT: thermostat_status
        })
        return self.status.status

    async def set_hot_water_ovrd(self, state, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Control the domestic hot water enable option. If the boiler has
        been configured to let the room unit control when to keep a
        small amount of water preheated, this command can influence
        that.
        @state should be 0 or 1 to enable the override in off or on
        state, or any other single character to disable the override.
        Return the accepted value, 'A' if the override is disabled
        or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_HOT_WATER
        status_otgw = {}
        ret = await self._wait_for_cmd(cmd, state, timeout)
        if ret is None:
            return None
        if ret == "A":
            status_otgw[v.OTGW_DHW_OVRD] = None
        elif ret in ["0", "1"]:
            ret = int(ret)
            status_otgw[v.OTGW_DHW_OVRD] = ret
        self.status.submit_partial_update(v.OTGW, status_otgw)
        return ret

    async def set_mode(self, mode, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Set the operating mode to either "Gateway" mode (:mode: =
        OTGW_MODE_GATEWAY or 1) or "Monitor" mode (:mode: =
        OTGW_MODE_MONITOR or 0), or use this method to reset the device
        (:mode: = OTGW_MODE_RESET).
        Return the newly activated mode, or the full renewed status
        dict after a reset.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_MODE
        status_otgw = {}
        ret = await self._wait_for_cmd(cmd, mode, timeout)
        if ret is None:
            return None
        if mode is v.OTGW_MODE_RESET:
            self.status.reset()
            await self.get_reports()
            await self.get_status()
            return self.status.status
        status_otgw[v.OTGW_MODE] = ret
        self.status.submit_partial_update(v.OTGW, status_otgw)
        return ret

    async def set_led_mode(self, led_id, mode, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Configure the functions of the six LEDs (A-F) that can
        optionally be connected to pins RB3/RB4/RB6/RB7 and the GPIO
        pins of the PIC. The following functions are currently
        available:

        R Receiving an Opentherm message from the thermostat or boiler
        X Transmitting an Opentherm message to the thermostat or boiler
        T Transmitting or receiving a message on the master interface
        B Transmitting or receiving a message on the slave interface
        O Remote setpoint override is active
        F Flame is on
        H Central heating is on
        W Hot water is on
        C Comfort mode (Domestic Hot Water Enable) is on
        E Transmission error has been detected
        M Boiler requires maintenance
        P Raised power mode active on thermostat interface.

        Return the new mode for the specified led, or None on failure.

        This method is a coroutine
        """
        if led_id in "ABCDEF" and mode in "RXTBOFHWCEMP":
            cmd = getattr(v, f"OTGW_CMD_LED_{led_id}")
            status_otgw = {}
            ret = await self._wait_for_cmd(cmd, mode, timeout)
            if ret is None:
                return None
            var = getattr(v, f"OTGW_LED_{led_id}")
            status_otgw[var] = ret
            self.status.submit_partial_update(v.OTGW, status_otgw)
            return ret

    async def set_gpio_mode(self, gpio_id, mode, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Configure the functions of the two GPIO pins of the gateway.
        The following functions are available:

        0 No function, default for both ports on a freshly flashed chip
        1 Ground - A permanently low output (0V). Could be used for a
            power LED
        2 Vcc - A permanently high output (5V). Can be used as a
            short-proof power supply for some external circuitry used
            by the other GPIO port
        3 LED E - An additional LED if you want to present more than 4
            LED functions
        4 LED F - An additional LED if you want to present more than 5
            LED functions
        5 Home - Set thermostat to setback temperature when pulled low
        6 Away - Set thermostat to setback temperature when pulled high
        7 DS1820 (GPIO port B only) - Data line for a DS18S20 or
            DS18B20 temperature sensor used to measure the outside
            temperature. A 4k7 resistor should be connected between
            GPIO port B and Vcc

        Return the new mode for the specified gpio, or None on
        failure.

        This method is a coroutine
        """
        if gpio_id in "AB" and mode in range(8):
            if mode == 7 and gpio_id != "B":
                return None
            cmd = getattr(v, f"OTGW_CMD_GPIO_{gpio_id}")
            status_otgw = {}
            ret = await self._wait_for_cmd(cmd, mode, timeout)
            if ret is None:
                return
            ret = int(ret)
            var = getattr(v, f"OTGW_GPIO_{gpio_id}")
            status_otgw[var] = ret
            self.status.submit_partial_update(v.OTGW, status_otgw)
            asyncio.ensure_future(self._poll_gpio())
            return ret

    async def set_setback_temp(self, sb_temp, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Configure the setback temperature to use in combination with
        GPIO functions HOME (5) and AWAY (6).
        Return the new setback temperature, or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_SETBACK
        status_otgw = {}
        ret = await self._wait_for_cmd(cmd, sb_temp, timeout)
        if ret is None:
            return
        ret = float(ret)
        status_otgw[v.OTGW_SB_TEMP] = ret
        self.status.submit_partial_update(v.OTGW, status_otgw)
        return ret

    async def add_alternative(self, alt, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Add the specified Data-ID to the list of alternative commands
        to send to the boiler instead of a Data-ID that is known to be
        unsupported by the boiler. Alternative Data-IDs will always be
        sent to the boiler in a Read-Data request message with the
        data-value set to zero. The table of alternative Data-IDs is
        stored in non-volatile memory so it will persist even if the
        gateway has been powered off. Data-ID values from 1 to 255 are
        allowed.
        Return the ID that was added to the list, or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_ADD_ALT
        alt = int(alt)
        if alt < 1 or alt > 255:
            return None
        ret = await self._wait_for_cmd(cmd, alt, timeout)
        if ret is not None:
            return int(ret)

    async def del_alternative(self, alt, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Remove the specified Data-ID from the list of alternative
        commands. Only one occurrence is deleted. If the Data-ID
        appears multiple times in the list of alternative commands,
        this command must be repeated to delete all occurrences. The
        table of alternative Data-IDs is stored in non-volatile memory
        so it will persist even if the gateway has been powered off.
        Data-ID values from 1 to 255 are allowed.
        Return the ID that was removed from the list, or None on
        failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_DEL_ALT
        alt = int(alt)
        if alt < 1 or alt > 255:
            return None
        ret = await self._wait_for_cmd(cmd, alt, timeout)
        if ret is not None:
            return int(ret)

    async def add_unknown_id(self, unknown_id, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Inform the gateway that the boiler doesn't support the
        specified Data-ID, even if the boiler doesn't indicate that
        by returning an Unknown-DataId response. Using this command
        allows the gateway to send an alternative Data-ID to the boiler
        instead.
        Return the added ID, or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_UNKNOWN_ID
        unknown_id = int(unknown_id)
        if unknown_id < 1 or unknown_id > 255:
            return None
        ret = await self._wait_for_cmd(cmd, unknown_id, timeout)
        if ret is not None:
            return int(ret)

    async def del_unknown_id(self, unknown_id, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Start forwarding the specified Data-ID to the boiler again.
        This command resets the counter used to determine if the
        specified Data-ID is supported by the boiler.
        Return the ID that was marked as supported, or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_KNOWN_ID
        unknown_id = int(unknown_id)
        if unknown_id < 1 or unknown_id > 255:
            return None
        ret = await self._wait_for_cmd(cmd, unknown_id, timeout)
        if ret is not None:
            return int(ret)

    async def set_max_ch_setpoint(self, temperature, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Set the maximum central heating setpoint. This command is only
        available with boilers that support this function.
        Return the newly accepted setpoint, or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_SET_MAX
        status_boiler = {}
        ret = await self._wait_for_cmd(cmd, temperature, timeout)
        if ret is None:
            return
        ret = float(ret)
        status_boiler[v.DATA_MAX_CH_SETPOINT] = ret
        self.status.submit_partial_update(v.BOILER, status_boiler)
        return ret

    async def set_dhw_setpoint(self, temperature, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Set the domestic hot water setpoint. This command is only
        available with boilers that support this function.
        Return the newly accepted setpoint, or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_SET_WATER
        status_boiler = {}
        ret = await self._wait_for_cmd(cmd, temperature, timeout)
        if ret is None:
            return
        ret = float(ret)
        status_boiler[v.DATA_DHW_SETPOINT] = ret
        self.status.submit_partial_update(v.BOILER, status_boiler)
        return ret

    async def set_max_relative_mod(self, max_mod, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Override the maximum relative modulation from the thermostat.
        Valid values are 0 through 100. Clear the setting by specifying
        a non-numeric value.
        Return the newly accepted value, '-' if a previous value was
        cleared, or None on failure.

        This method is a coroutine
        """
        if isinstance(max_mod, int) and not 0 <= max_mod <= 100:
            return None
        cmd = v.OTGW_CMD_MAX_MOD
        status_boiler = {}
        ret = await self._wait_for_cmd(cmd, max_mod, timeout)
        if ret is None:
            return
        if ret == "-":
            status_boiler[v.DATA_SLAVE_MAX_RELATIVE_MOD] = None
        else:
            ret = int(ret)
            status_boiler[v.DATA_SLAVE_MAX_RELATIVE_MOD] = ret
        self.status.submit_partial_update(v.BOILER, status_boiler)
        return ret

    async def set_control_setpoint(self, setpoint, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Manipulate the control setpoint being sent to the boiler. Set
        to 0 to pass along the value specified by the thermostat.
        Return the newly accepted value, or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_CONTROL_SETPOINT
        status_boiler = {}
        ret = await self._wait_for_cmd(cmd, setpoint, timeout)
        if ret is None:
            return
        ret = float(ret)
        status_boiler[v.DATA_CONTROL_SETPOINT] = ret
        self.status.submit_partial_update(v.BOILER, status_boiler)
        return ret

    async def set_control_setpoint_2(self, setpoint, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Manipulate the control setpoint being sent to the boiler for the second
        heating circuit. Set to 0 to pass along the value specified by the thermostat.
        Return the newly accepted value, or None on failure.

        This method is a coroutine
        """
        cmd = v.OTGW_CMD_CONTROL_SETPOINT_2
        status_boiler = {}
        ret = await self._wait_for_cmd(cmd, setpoint, timeout)
        if ret is None:
            return
        ret = float(ret)
        status_boiler[v.DATA_CONTROL_SETPOINT_2] = ret
        self.status.submit_partial_update(v.BOILER, status_boiler)
        return ret

    async def set_ch_enable_bit(self, ch_bit, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Control the CH enable status bit when overriding the control
        setpoint. By default the CH enable bit is set after a call to
        set_control_setpoint with a value other than 0. With this
        method, the bit can be manipulated.
        @ch_bit can be either 0 or 1.
        Return the newly accepted value (0 or 1), or None on failure.

        This method is a coroutine
        """
        if ch_bit not in [0, 1]:
            return None
        cmd = v.OTGW_CMD_CONTROL_HEATING
        status_boiler = {}
        ret = await self._wait_for_cmd(cmd, ch_bit, timeout)
        if ret is None:
            return
        ret = int(ret)
        status_boiler[v.DATA_MASTER_CH_ENABLED] = ret
        self.status.submit_partial_update(v.BOILER, status_boiler)
        return ret

    async def set_ch2_enable_bit(self, ch_bit, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Control the CH enable status bit when overriding the control
        setpoint. By default the CH enable bit is set after a call to
        set_control_setpoint with a value other than 0. With this
        method, the bit can be manipulated.
        @ch_bit can be either 0 or 1.
        Return the newly accepted value (0 or 1), or None on failure.

        This method is a coroutine
        """
        if ch_bit not in [0, 1]:
            return None
        cmd = v.OTGW_CMD_CONTROL_HEATING_2
        status_boiler = {}
        ret = await self._wait_for_cmd(cmd, ch_bit, timeout)
        if ret is None:
            return
        ret = int(ret)
        status_boiler[v.DATA_MASTER_CH2_ENABLED] = ret
        self.status.submit_partial_update(v.BOILER, status_boiler)
        return ret

    async def set_ventilation(self, pct, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Configure a ventilation setpoint override value (0-100%).
        Return the newly accepted value, or None on failure.
        @pct :int: Must be between 0 and 100.

        This method is a coroutine
        """
        if not 0 <= pct <= 100:
            return None
        cmd = v.OTGW_CMD_VENT
        status_boiler = {}
        ret = await self._wait_for_cmd(cmd, pct, timeout)
        if ret is None:
            return
        ret = int(ret)
        status_boiler[v.DATA_COOLING_CONTROL] = ret
        self.status.submit_partial_update(v.BOILER, status_boiler)
        return ret

    async def send_transparent_command(
        self, cmd, state, timeout=v.OTGW_DEFAULT_TIMEOUT
    ):
        """
        Sends custom otgw commands throug a transparent interface.
        Check https://otgw.tclcode.com/firmware.html for supported commands.
        @cmd the supported command e.g. 'SC' (set time/day)
        @state the command argument e.g. '23:59/4' (the current time/day)
        Returns the gateway response, which should be equal to @state.

        Note that unlike the set_* methods, this does not update the status object.

        This method is a coroutine
        """
        ret = await self._wait_for_cmd(cmd, state, timeout)
        return ret

    def subscribe(self, coro):
        """
        Subscribe to status updates from the Opentherm Gateway.
        Can only be used after connect()
        @coro is a coroutine which will be called with a single
        argument (status) when a status change occurs.
        Return True on success, False if not connected or already
        subscribed.
        """
        return self.status.subscribe(coro)

    def unsubscribe(self, coro):
        """
        Unsubscribe from status updates from the Opentherm Gateway.
        Can only be used after connect()
        @coro is a coroutine which has been subscribed with subscribe()
        earlier.
        Return True on success, false if not connected or subscribed.
        """
        return self.status.unsubscribe(coro)

    async def _wait_for_cmd(self, cmd, value, timeout=v.OTGW_DEFAULT_TIMEOUT):
        """
        Wrap @cmd in applicable asyncio call.

        This method is a coroutine.
        """
        if not self.connection.connected:
            return None
        try:
            return await asyncio.wait_for(
                self._protocol.command_processor.issue_cmd(cmd, value),
                timeout,
            )
        except asyncio.TimeoutError:
            _LOGGER.error("Timed out waiting for command: %s, value: %s.", cmd, value)
            return
        except (RuntimeError, SyntaxError, ValueError) as exc:
            _LOGGER.error(
                "Command %s with value %s raised exception: %s", cmd, value, exc
            )

    async def _poll_gpio(self, interval=10):
        """
        Start or stop polling GPIO states.

        GPIO states aren't being pushed by the gateway, we need to poll
        if we want updates.
        """
        poll = 0 in (
            self.status.status[v.OTGW].get(v.OTGW_GPIO_A),
            self.status.status[v.OTGW].get(v.OTGW_GPIO_B),
        )
        if poll and self._gpio_task is None:

            async def polling_routine():
                """Poll GPIO state every @interval seconds."""
                try:
                    while True:
                        ret = await self._wait_for_cmd(
                            v.OTGW_CMD_REPORT, v.OTGW_REPORT_GPIO_STATES
                        )
                        if ret:
                            pios = ret[2:]
                            status_otgw = {
                                v.OTGW_GPIO_A_STATE: int(pios[0]),
                                v.OTGW_GPIO_B_STATE: int(pios[1]),
                            }
                            self.status.submit_partial_update(v.OTGW, status_otgw)
                        await asyncio.sleep(interval)
                except asyncio.CancelledError:
                    status_otgw = {
                        v.OTGW_GPIO_A_STATE: 0,
                        v.OTGW_GPIO_B_STATE: 0,
                    }
                    self.status.submit_partial_update(v.OTGW, status_otgw)
                    self._gpio_task = None
                    _LOGGER.debug("GPIO polling routine stopped")

            _LOGGER.debug("Starting GPIO polling routine")
            self._gpio_task = asyncio.get_running_loop().create_task(polling_routine())
        elif not poll and self._gpio_task is not None:
            _LOGGER.debug("Stopping GPIO polling routine")
            self._gpio_task.cancel()


def process_statusfields_v4(status_fields):
    """
    Process the fields of a split status line for OpenTherm Gateway firmware
    version <5.0.
    
    Return a tuple of (boiler_status, thermostat_status).
    """
    device_status = status_fields[0].split("/")
    master_status = device_status[0]
    slave_status = device_status[1]
    remote_params = status_fields[2].split("/")
    capmodlimits = status_fields[4].split("/")
    dhw_setp_bounds = status_fields[13].split("/")
    ch_setp_bounds = status_fields[14].split("/")
    thermostat_status = {
        v.DATA_MASTER_CH_ENABLED: int(master_status[7]),
        v.DATA_MASTER_DHW_ENABLED: int(master_status[6]),
        v.DATA_MASTER_COOLING_ENABLED: int(master_status[5]),
        v.DATA_MASTER_OTC_ENABLED: int(master_status[4]),
        v.DATA_MASTER_CH2_ENABLED: int(master_status[3]),
        v.DATA_CONTROL_SETPOINT: float(status_fields[1]),
        v.DATA_ROOM_SETPOINT: float(status_fields[5]),
        v.DATA_ROOM_TEMP: float(status_fields[8]),
    }
    boiler_status = {
        v.DATA_SLAVE_FAULT_IND: int(slave_status[7]),
        v.DATA_SLAVE_CH_ACTIVE: int(slave_status[6]),
        v.DATA_SLAVE_DHW_ACTIVE: int(slave_status[5]),
        v.DATA_SLAVE_FLAME_ON: int(slave_status[4]),
        v.DATA_SLAVE_COOLING_ACTIVE: int(slave_status[3]),
        v.DATA_SLAVE_CH2_ACTIVE: int(slave_status[2]),
        v.DATA_SLAVE_DIAG_IND: int(slave_status[1]),
        v.DATA_REMOTE_TRANSFER_DHW: int(remote_params[0][7]),
        v.DATA_REMOTE_TRANSFER_MAX_CH: int(remote_params[0][6]),
        v.DATA_REMOTE_RW_DHW: int(remote_params[1][7]),
        v.DATA_REMOTE_RW_MAX_CH: int(remote_params[1][6]),
        v.DATA_SLAVE_MAX_RELATIVE_MOD: float(status_fields[3]),
        v.DATA_SLAVE_MAX_CAPACITY: int(capmodlimits[0]),
        v.DATA_SLAVE_MIN_MOD_LEVEL: int(capmodlimits[1]),
        v.DATA_REL_MOD_LEVEL: float(status_fields[6]),
        v.DATA_CH_WATER_PRESS: float(status_fields[7]),
        v.DATA_CH_WATER_TEMP: float(status_fields[9]),
        v.DATA_DHW_TEMP: float(status_fields[10]),
        v.DATA_OUTSIDE_TEMP: float(status_fields[11]),
        v.DATA_RETURN_WATER_TEMP: float(status_fields[12]),
        v.DATA_SLAVE_DHW_MAX_SETP: int(dhw_setp_bounds[0]),
        v.DATA_SLAVE_DHW_MIN_SETP: int(dhw_setp_bounds[1]),
        v.DATA_SLAVE_CH_MAX_SETP: int(ch_setp_bounds[0]),
        v.DATA_SLAVE_CH_MIN_SETP: int(ch_setp_bounds[1]),
        v.DATA_DHW_SETPOINT: float(status_fields[15]),
        v.DATA_MAX_CH_SETPOINT: float(status_fields[16]),
        v.DATA_TOTAL_BURNER_STARTS: int(status_fields[17]),
        v.DATA_CH_PUMP_STARTS: int(status_fields[18]),
        v.DATA_DHW_PUMP_STARTS: int(status_fields[19]),
        v.DATA_DHW_BURNER_STARTS: int(status_fields[20]),
        v.DATA_TOTAL_BURNER_HOURS: int(status_fields[21]),
        v.DATA_CH_PUMP_HOURS: int(status_fields[22]),
        v.DATA_DHW_PUMP_HOURS: int(status_fields[23]),
        v.DATA_DHW_BURNER_HOURS: int(status_fields[24]),
    }
    return (boiler_status, thermostat_status)


def process_statusfields_v5(status_fields):
    """
    Process the fields of a split status line for OpenTherm Gateway firmware
    version >=5.0.
    
    Return a tuple of (boiler_status, thermostat_status).
    """
    device_status = status_fields[0].split("/")
    master_status = device_status[0]
    slave_status = device_status[1]
    remote_params = status_fields[2].split("/")
    capmodlimits = status_fields[6].split("/")
    dhw_setp_bounds = status_fields[19].split("/")
    ch_setp_bounds = status_fields[20].split("/")
    vh_device_status = status_fields[23].split("/")
    vh_master_status = vh_device_status[0]
    vh_slave_status = vh_device_status[1]
    thermostat_status = {
        v.DATA_MASTER_CH_ENABLED: int(master_status[7]),
        v.DATA_MASTER_DHW_ENABLED: int(master_status[6]),
        v.DATA_MASTER_COOLING_ENABLED: int(master_status[5]),
        v.DATA_MASTER_OTC_ENABLED: int(master_status[4]),
        v.DATA_MASTER_CH2_ENABLED: int(master_status[3]),
        v.DATA_CONTROL_SETPOINT: float(status_fields[1]),
        v.DATA_ROOM_SETPOINT: float(status_fields[7]),
        v.DATA_COOLING_CONTROL: float(status_fields[3]),
        v.DATA_CONTROL_SETPOINT_2: float(status_fields[4]),
        v.DATA_ROOM_SETPOINT_2: float(status_fields[11]),
        v.DATA_ROOM_TEMP: float(status_fields[12]),
        v.DATA_VH_MASTER_VENT_ENABLED: int(vh_master_status[7]),
        v.DATA_VH_MASTER_BYPASS_POS: int(vh_master_status[6]),
        v.DATA_VH_MASTER_BYPASS_MODE: int(vh_master_status[5]),
        v.DATA_VH_MASTER_FREE_VENT_MODE: int(vh_master_status[4]),
        v.DATA_VH_CONTROL_SETPOINT: int(status_fields[24]),
    }
    boiler_status = {
        v.DATA_SLAVE_FAULT_IND: int(slave_status[7]),
        v.DATA_SLAVE_CH_ACTIVE: int(slave_status[6]),
        v.DATA_SLAVE_DHW_ACTIVE: int(slave_status[5]),
        v.DATA_SLAVE_FLAME_ON: int(slave_status[4]),
        v.DATA_SLAVE_COOLING_ACTIVE: int(slave_status[3]),
        v.DATA_SLAVE_CH2_ACTIVE: int(slave_status[2]),
        v.DATA_SLAVE_DIAG_IND: int(slave_status[1]),
        v.DATA_REMOTE_TRANSFER_DHW: int(remote_params[0][7]),
        v.DATA_REMOTE_TRANSFER_MAX_CH: int(remote_params[0][6]),
        v.DATA_REMOTE_RW_DHW: int(remote_params[1][7]),
        v.DATA_REMOTE_RW_MAX_CH: int(remote_params[1][6]),
        v.DATA_SLAVE_MAX_RELATIVE_MOD: float(status_fields[5]),
        v.DATA_SLAVE_MAX_CAPACITY: int(capmodlimits[0]),
        v.DATA_SLAVE_MIN_MOD_LEVEL: int(capmodlimits[1]),
        v.DATA_REL_MOD_LEVEL: float(status_fields[8]),
        v.DATA_CH_WATER_PRESS: float(status_fields[9]),
        v.DATA_DHW_FLOW_RATE: float(status_fields[10]),
        v.DATA_CH_WATER_TEMP: float(status_fields[13]),
        v.DATA_DHW_TEMP: float(status_fields[14]),
        v.DATA_OUTSIDE_TEMP: float(status_fields[15]),
        v.DATA_RETURN_WATER_TEMP: float(status_fields[16]),
        v.DATA_CH_WATER_TEMP_2: float(status_fields[17]),
        v.DATA_EXHAUST_TEMP: int(status_fields[18]),
        v.DATA_SLAVE_DHW_MAX_SETP: int(dhw_setp_bounds[0]),
        v.DATA_SLAVE_DHW_MIN_SETP: int(dhw_setp_bounds[1]),
        v.DATA_SLAVE_CH_MAX_SETP: int(ch_setp_bounds[0]),
        v.DATA_SLAVE_CH_MIN_SETP: int(ch_setp_bounds[1]),
        v.DATA_DHW_SETPOINT: float(status_fields[21]),
        v.DATA_MAX_CH_SETPOINT: float(status_fields[22]),
        v.DATA_VH_SLAVE_FAULT_INDICATE: int(vh_slave_status[7]),
        v.DATA_VH_SLAVE_VENT_MODE: int(vh_slave_status[6]),
        v.DATA_VH_SLAVE_BYPASS_STATUS: int(vh_slave_status[5]),
        v.DATA_VH_SLAVE_BYPASS_AUTO_STATUS: int(vh_slave_status[4]),
        v.DATA_VH_SLAVE_FREE_VENT_STATUS: int(vh_slave_status[3]),
        v.DATA_VH_SLAVE_DIAG_INDICATE: int(vh_slave_status[1]),
        v.DATA_TOTAL_BURNER_STARTS: int(status_fields[26]),
        v.DATA_CH_PUMP_STARTS: int(status_fields[27]),
        v.DATA_DHW_PUMP_STARTS: int(status_fields[28]),
        v.DATA_DHW_BURNER_STARTS: int(status_fields[29]),
        v.DATA_TOTAL_BURNER_HOURS: int(status_fields[30]),
        v.DATA_CH_PUMP_HOURS: int(status_fields[31]),
        v.DATA_DHW_PUMP_HOURS: int(status_fields[32]),
        v.DATA_DHW_BURNER_HOURS: int(status_fields[33]),
        v.DATA_VH_RELATIVE_VENT: int(status_fields[25]),
    }
    return (boiler_status, thermostat_status)



"""Config and fixtures for pyotgw tests"""

import asyncio
from unittest.mock import MagicMock, patch

import pytest
import pytest_asyncio

import pyotgw
from pyotgw.connection import ConnectionManager, ConnectionWatchdog
from pyotgw.status import StatusManager


@pytest_asyncio.fixture
async def pygw():
    """Return a basic pyotgw object"""
    gw = pyotgw.OpenThermGateway()
    await gw.connection.watchdog.stop()
    yield gw
    await gw.cleanup()


@pytest_asyncio.fixture
async def pygw_proto(pygw):
    """Return a "connected" protocol object"""

    async def empty_coroutine():
        return

    trans = MagicMock(loop=asyncio.get_running_loop())
    activity_callback = MagicMock(side_effect=empty_coroutine)
    proto = pyotgw.protocol.OpenThermProtocol(
        pygw.status,
        activity_callback,
    )
    proto.activity_callback = activity_callback
    pygw._transport = trans
    pygw._protocol = proto
    with patch(
        "pyotgw.messageprocessor.MessageProcessor._process_msgs", return_value=None
    ):
        proto.connection_made(trans)
    yield proto
    await proto.cleanup()


@pytest_asyncio.fixture
async def pygw_status():
    """Return a StatusManager object"""
    status_manager = StatusManager()
    yield status_manager
    await status_manager.cleanup()


@pytest.fixture
def pygw_message_processor(pygw_proto):
    """Return a MessageProcessor object"""
    return pygw_proto.message_processor


@pytest_asyncio.fixture
async def pygw_conn(pygw):
    """Return a ConnectionManager object"""
    connection_manager = ConnectionManager(pygw)
    yield connection_manager
    await connection_manager.watchdog.stop()


@pytest_asyncio.fixture
async def pygw_watchdog():
    """Return a ConnectionWatchdog object"""
    watchdog = ConnectionWatchdog()
    yield watchdog
    await watchdog.stop()


@pytest_asyncio.fixture(autouse=True)
async def check_task_cleanup():
    loop = asyncio.get_running_loop()
    task_count = len(asyncio.all_tasks(loop))

    yield

    assert len(asyncio.all_tasks(loop)) == task_count, "Test is leaving tasks behind!"



"""Test for pyotgw/messageprocessor.py"""

import asyncio
import logging
import re
from unittest.mock import MagicMock, patch

import pytest

from pyotgw import vars as v
from tests.data import pygw_proto_messages
from tests.helpers import called_once

MATCH_PATTERN = r"^(T|B|R|A|E)([0-9A-F]{8})$"


@pytest.mark.asyncio
async def test_cleanup(pygw_message_processor):
    """Test MessageProcessor.cleanup()"""
    assert pygw_message_processor._task
    await pygw_message_processor.cleanup()
    assert not pygw_message_processor._task


def test_connection_lost(pygw_message_processor):
    """Test MessageProcessor.connection_lost()"""
    message = re.match(MATCH_PATTERN, "A01020304")
    pygw_message_processor.submit_matched_message(message)
    pygw_message_processor.submit_matched_message(message)
    pygw_message_processor.submit_matched_message(message)
    pygw_message_processor.connection_lost()
    assert pygw_message_processor._msgq.empty()


def test_submit_matched_message(caplog, pygw_message_processor):
    """Tests MessageProcessor.submit_matched_message()"""
    bad_match = re.match(MATCH_PATTERN, "E01020304")
    good_match = re.match(MATCH_PATTERN, "A01020304")

    pygw_message_processor.submit_matched_message(bad_match)
    assert pygw_message_processor._msgq.empty()

    with caplog.at_level(logging.DEBUG):
        pygw_message_processor.submit_matched_message(good_match)
    assert caplog.record_tuples == [
        (
            "pyotgw.messageprocessor",
            logging.DEBUG,
            "Added line to message queue. Queue size: 1",
        ),
    ]
    assert pygw_message_processor._msgq.get_nowait() == (
        "A",
        v.READ_DATA,
        b"\x02",
        b"\x03",
        b"\x04",
    )


def test_dissect_msg(caplog, pygw_message_processor):
    """Test MessageProcessor._dissect_msg"""
    pat = r"^(T|B|R|A|E)([0-9A-F]{8})$"
    test_matches = (
        re.match(pat, "A10203040"),
        re.match(pat, "EEEEEEEEE"),
        re.match(pat, "AEEEEEEEE"),
    )
    none_tuple = (None, None, None, None, None)

    assert pygw_message_processor._dissect_msg(test_matches[0]) == (
        "A",
        v.WRITE_DATA,
        b"\x20",
        b"\x30",
        b"\x40",
    )

    with caplog.at_level(logging.INFO):
        assert pygw_message_processor._dissect_msg(test_matches[1]) == none_tuple

    assert caplog.record_tuples == [
        (
            "pyotgw.messageprocessor",
            logging.INFO,
            "The OpenTherm Gateway received an erroneous message."
            " This is not a bug in pyotgw. Ignoring: EEEEEEEE",
        )
    ]

    assert pygw_message_processor._dissect_msg(test_matches[2]) == none_tuple


def test_get_msgtype(pygw_message_processor):
    """Test MessageProcessor._get_msgtype()"""
    assert pygw_message_processor._get_msgtype(int("11011111", 2)) == int("0101", 2)
    assert pygw_message_processor._get_msgtype(int("01000001", 2)) == int("0100", 2)


@pytest.mark.asyncio
async def test_process_msgs(caplog, pygw_message_processor):
    """Test MessageProcessor._process_msgs()"""
    test_case = (
        "B",
        v.READ_ACK,
        b"\x23",
        b"\x0a",
        b"\x01",
    )
    with patch.object(
        pygw_message_processor, "_process_msg"
    ) as process_msg, caplog.at_level(logging.DEBUG):
        task = asyncio.create_task(pygw_message_processor._process_msgs())
        pygw_message_processor._msgq.put_nowait(test_case)
        await called_once(process_msg)

    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task

    process_msg.assert_called_once_with(test_case)
    assert caplog.record_tuples == [
        ("pyotgw.messageprocessor", logging.DEBUG, "Processing: B 04 23 0A 01")
    ]


@pytest.mark.asyncio
async def test_process_msg(pygw_message_processor):
    """Test MessageProcessor._process_msg()"""
    # Test quirks
    test_case = (
        "B",
        v.READ_ACK,
        v.MSG_TROVRD,
        b"\x10",
        b"\x80",
    )

    with patch.object(
        pygw_message_processor, "_quirk_trovrd", return_value=None
    ) as quirk_trovrd:
        await pygw_message_processor._process_msg(test_case)

    quirk_trovrd.assert_called_once_with(
        v.BOILER,
        "B",
        b"\x10",
        b"\x80",
    )

    async def empty_coroutine(status):
        return

    status_callback = MagicMock(side_effect=empty_coroutine)
    pygw_message_processor.status_manager.subscribe(status_callback)

    for test_case, expected_result in pygw_proto_messages:
        pygw_message_processor.status_manager.reset()
        pygw_message_processor.status_manager._updateq.get_nowait()
        await pygw_message_processor._process_msg(test_case)
        if expected_result is not None:
            await called_once(status_callback)
            status_callback.assert_called_once_with(expected_result)
            status_callback.reset_mock()


@pytest.mark.asyncio
async def test_get_dict_update_for_action():
    """Test MessageProcessor._get_dict_update_for_action"""
    assert True  # Fully tested in test_process_msg()


@pytest.mark.asyncio
async def test_quirk_trovrd(pygw_message_processor):
    """Test MessageProcessor._quirk_trovrd()"""

    async def empty_coroutine(stat):
        return

    status_callback = MagicMock(side_effect=empty_coroutine)
    pygw_message_processor.status_manager.subscribe(status_callback)
    pygw_message_processor.status_manager.submit_partial_update(
        v.OTGW,
        {v.OTGW_THRM_DETECT: "I"},
    )
    await called_once(status_callback)
    status_callback.reset_mock()

    with patch.object(
        pygw_message_processor.command_processor,
        "issue_cmd",
        return_value="O=c19.5",
    ):
        await pygw_message_processor._quirk_trovrd(
            v.THERMOSTAT,
            "A",
            b"\x15",
            b"\x40",
        )

    await called_once(status_callback)
    status_callback.assert_called_once_with(
        {
            v.BOILER: {},
            v.OTGW: {v.OTGW_THRM_DETECT: "I"},
            v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: 19.5},
        }
    )

    with patch.object(
        pygw_message_processor.command_processor,
        "issue_cmd",
        return_value="O=q---",
    ), patch.object(
        pygw_message_processor.status_manager,
        "submit_partial_update",
    ) as partial_update, patch.object(
        pygw_message_processor.status_manager,
        "delete_value",
    ) as delete_value:
        await pygw_message_processor._quirk_trovrd(
            v.THERMOSTAT,
            "A",
            b"\x15",
            b"\x40",
        )

    partial_update.assert_not_called()
    delete_value.assert_not_called()
    assert (
        v.DATA_ROOM_SETPOINT_OVRD
        in pygw_message_processor.status_manager.status[v.THERMOSTAT]
    )

    status_callback.reset_mock()
    await pygw_message_processor._quirk_trovrd(
        v.THERMOSTAT,
        "A",
        b"\x00",
        b"\x00",
    )
    await called_once(status_callback)
    status_callback.assert_called_once_with(
        {
            v.BOILER: {},
            v.OTGW: {v.OTGW_THRM_DETECT: "I"},
            v.THERMOSTAT: {},
        }
    )

    status_callback.reset_mock()
    pygw_message_processor.status_manager.submit_partial_update(
        v.OTGW, {v.OTGW_THRM_DETECT: "D"}
    )
    await called_once(status_callback)
    status_callback.reset_mock()

    await pygw_message_processor._quirk_trovrd(
        v.THERMOSTAT,
        "A",
        b"\x15",
        b"\x40",
    )
    await called_once(status_callback)
    status_callback.assert_called_once_with(
        {
            v.BOILER: {},
            v.OTGW: {v.OTGW_THRM_DETECT: "D"},
            v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: 21.25},
        }
    )

    status_callback.reset_mock()
    pygw_message_processor.status_manager.submit_partial_update(
        v.OTGW, {v.OTGW_THRM_DETECT: "I"}
    )
    await called_once(status_callback)
    status_callback.reset_mock()

    with patch.object(
        pygw_message_processor.command_processor,
        "issue_cmd",
        return_value="O=N",
    ):
        await pygw_message_processor._quirk_trovrd(
            v.THERMOSTAT,
            "A",
            b"\x15",
            b"\x40",
        )
    await called_once(status_callback)
    status_callback.assert_called_once_with(
        {
            v.BOILER: {},
            v.OTGW: {v.OTGW_THRM_DETECT: "I"},
            v.THERMOSTAT: {},
        }
    )


@pytest.mark.asyncio
async def test_quirk_trset_s2m(pygw_message_processor):
    """Test MessageProcessor._quirk_trset_s2m()"""

    async def empty_coroutine(stat):
        return

    with patch.object(
        pygw_message_processor.status_manager, "submit_partial_update"
    ) as partial_update:
        await pygw_message_processor._quirk_trset_s2m(
            v.THERMOSTAT,
            b"\x01",
            b"\x02",
        )
        await pygw_message_processor._quirk_trset_s2m(v.BOILER, b"\x14", b"\x80")

    partial_update.assert_called_once_with(v.BOILER, {v.DATA_ROOM_SETPOINT: 20.5})


def test_get_flag8(pygw_message_processor):
    """Test pygw._get_flag8()"""
    test_cases = (
        (
            int("00000001", 2).to_bytes(1, "big"),
            [1, 0, 0, 0, 0, 0, 0, 0],
        ),
        (
            int("00000010", 2).to_bytes(1, "big"),
            [0, 1, 0, 0, 0, 0, 0, 0],
        ),
        (
            int("00000100", 2).to_bytes(1, "big"),
            [0, 0, 1, 0, 0, 0, 0, 0],
        ),
        (
            int("00001000", 2).to_bytes(1, "big"),
            [0, 0, 0, 1, 0, 0, 0, 0],
        ),
        (
            int("00010000", 2).to_bytes(1, "big"),
            [0, 0, 0, 0, 1, 0, 0, 0],
        ),
        (
            int("00100000", 2).to_bytes(1, "big"),
            [0, 0, 0, 0, 0, 1, 0, 0],
        ),
        (
            int("01000000", 2).to_bytes(1, "big"),
            [0, 0, 0, 0, 0, 0, 1, 0],
        ),
        (
            int("10000000", 2).to_bytes(1, "big"),
            [0, 0, 0, 0, 0, 0, 0, 1],
        ),
    )

    for case, res in test_cases:
        assert pygw_message_processor._get_flag8(case) == res


def test_get_u8(pygw_message_processor):
    """Test pygw._get_u8()"""
    test_cases = (
        (
            b"\x00",
            0,
        ),
        (
            b"\xff",
            255,
        ),
    )

    for case, res in test_cases:
        assert pygw_message_processor._get_u8(case) == res


def test_get_s8(pygw_message_processor):
    """Test pygw._get_s8()"""
    test_cases = (
        (
            b"\x00",
            0,
        ),
        (
            b"\xff",
            -1,
        ),
    )

    for case, res in test_cases:
        assert pygw_message_processor._get_s8(case) == res


def test_get_f8_8(pygw_message_processor):
    """Test pygw._get_f8_8()"""
    test_cases = (
        (
            (
                b"\x00",
                b"\x00",
            ),
            0.0,
        ),
        (
            (
                b"\xff",
                b"\x80",
            ),
            -0.5,
        ),
    )

    for case, res in test_cases:
        assert pygw_message_processor._get_f8_8(*case) == res


def test_get_u16(pygw_message_processor):
    """Test pygw._get_u16()"""
    test_cases = (
        (
            (
                b"\x00",
                b"\x00",
            ),
            0,
        ),
        (
            (
                b"\xff",
                b"\xff",
            ),
            65535,
        ),
    )

    for case, res in test_cases:
        assert pygw_message_processor._get_u16(*case) == res


def test_get_s16(pygw_message_processor):
    """Test pygw._get_s16()"""
    test_cases = (
        (
            (
                b"\x00",
                b"\x00",
            ),
            0,
        ),
        (
            (
                b"\xff",
                b"\xff",
            ),
            -1,
        ),
    )

    for case, res in test_cases:
        assert pygw_message_processor._get_s16(*case) == res



"""Tests for pyotgw/pyotgw.py"""

import asyncio
import logging
from datetime import datetime
from unittest.mock import MagicMock, call, patch

import pytest
import serial

import pyotgw.vars as v
from tests.data import pygw_reports, pygw_status
from tests.helpers import called_once, called_x_times


@pytest.mark.asyncio
async def test_cleanup(pygw):
    """Test pyotgw.cleanup()"""
    pygw.status.submit_partial_update(v.OTGW, {v.OTGW_GPIO_A: 0})
    pygw.loop = asyncio.get_running_loop()

    with patch.object(pygw, "_wait_for_cmd"):
        await pygw._poll_gpio()
        assert pygw._gpio_task
        await pygw.cleanup()
        assert not pygw._gpio_task


@pytest.mark.asyncio
async def test_connect_success_and_reconnect_with_gpio(caplog, pygw, pygw_proto):
    """Test pyotgw.connect()"""
    with patch.object(pygw, "get_reports", return_value={}), patch.object(
        pygw,
        "get_status",
        return_value={},
    ), patch.object(pygw, "_poll_gpio") as poll_gpio, patch.object(
        pygw_proto,
        "init_and_wait_for_activity",
    ) as init_and_wait, patch(
        "serial_asyncio_fast.create_serial_connection",
        return_value=(pygw_proto.transport, pygw_proto),
    ), caplog.at_level(
        logging.DEBUG
    ):
        status = await pygw.connect("loop://")

        assert status == v.DEFAULT_STATUS
        init_and_wait.assert_called_once()
        poll_gpio.assert_called_once()

        await pygw.connection.watchdog.stop()
        await pygw.connection.watchdog._callback()

        assert (
            "pyotgw.connection",
            logging.DEBUG,
            "Scheduling reconnect...",
        ) in caplog.record_tuples

        await pygw.disconnect()


@pytest.mark.asyncio
async def test_connect_skip_init(caplog, pygw, pygw_proto):
    """Test pyotgw.connect() with skip_init"""
    with patch.object(
        pygw, "get_reports", return_value={}
    ) as get_reports, patch.object(
        pygw,
        "get_status",
        return_value={},
    ) as get_status, patch.object(
        pygw, "_poll_gpio"
    ) as poll_gpio, patch.object(
        pygw_proto,
        "init_and_wait_for_activity",
    ) as init_and_wait, patch(
        "serial_asyncio_fast.create_serial_connection",
        return_value=(pygw_proto.transport, pygw_proto),
    ), caplog.at_level(
        logging.DEBUG
    ):
        status = await pygw.connect("loop://", skip_init=True)

        assert status == v.DEFAULT_STATUS
        init_and_wait.assert_called_once()
        poll_gpio.assert_called_once()
        get_reports.assert_not_awaited()
        get_status.assert_not_awaited()

        await pygw.connection.watchdog.stop()
        await pygw.connection.watchdog._callback()

        assert (
            "pyotgw.connection",
            logging.DEBUG,
            "Scheduling reconnect...",
        ) in caplog.record_tuples

        await pygw.disconnect()


@pytest.mark.asyncio
async def test_connect_serialexception(caplog, pygw):
    """Test pyotgw.connect() with SerialException"""
    loop = asyncio.get_running_loop()

    with patch(
        "serial_asyncio_fast.create_serial_connection",
        side_effect=serial.serialutil.SerialException,
    ) as create_serial_connection, patch.object(
        pygw.connection,
        "_get_retry_timeout",
        return_value=0,
    ) as loops_done:
        task = loop.create_task(pygw.connect("loop://"))

        await called_x_times(loops_done, 2)

        assert isinstance(pygw.connection._connecting_task, asyncio.Task)
        assert len(caplog.records) == 1
        assert caplog.record_tuples == [
            (
                "pyotgw.connection",
                logging.ERROR,
                "Could not connect to serial device on loop://. "
                "Will keep trying. Reported error was: ",
            ),
        ]
        assert create_serial_connection.call_count > 1

        task.cancel()
        assert await task is False


@pytest.mark.asyncio
async def test_connect_cancel(pygw):
    """Test pyotgw.connect() with CancelledError"""
    with patch(
        "serial_asyncio_fast.create_serial_connection",
        side_effect=asyncio.CancelledError,
    ) as create_serial_connection:
        status = await pygw.connect("loop://")

    assert status is False
    create_serial_connection.assert_called_once()


@pytest.mark.asyncio
async def test_connect_timeouterror(caplog, pygw, pygw_proto):
    """Test pyotgw.connect() with TimeoutError"""
    loop = asyncio.get_running_loop()

    # Mock these before the 'with' context manager to ensure the mocks get
    # included in the patched response for create_serial_connection().
    pygw_proto.init_and_wait_for_activity = MagicMock(side_effect=asyncio.TimeoutError)
    pygw_proto.disconnect = MagicMock()

    with patch.object(
        pygw.connection,
        "_get_retry_timeout",
        return_value=0,
    ) as loops_done, patch(
        "serial_asyncio_fast.create_serial_connection",
        return_value=(pygw_proto.transport, pygw_proto),
    ), caplog.at_level(
        logging.DEBUG
    ):
        task = loop.create_task(pygw.connect("loop://"))
        await called_x_times(loops_done, 2)

        assert isinstance(pygw.connection._connecting_task, asyncio.Task)
        assert len(caplog.records) == 1
        assert caplog.record_tuples == [
            (
                "pyotgw.connection",
                logging.ERROR,
                "The serial device on loop:// is not responding. " "Will keep trying.",
            ),
        ]
        assert pygw_proto.init_and_wait_for_activity.call_count > 1
        assert pygw_proto.disconnect.call_count > 1

        task.cancel()
        assert await task is False


@pytest.mark.asyncio
async def test_disconnect_while_connecting(pygw):
    """Test pyotgw.disconnect()"""
    loop = asyncio.get_running_loop()

    with patch(
        "serial_asyncio_fast.create_serial_connection",
        side_effect=serial.SerialException,
    ):
        task = loop.create_task(pygw.connect("loop://"))

        with patch(
            "pyotgw.protocol.OpenThermProtocol.disconnect"
        ) as protocol_disconnect:
            await pygw.disconnect()
        assert await task is False
    assert not protocol_disconnect.called


def test_set_connection_options(pygw):
    """Test pyotgw.set_connection_options"""
    # Just test forwarding, we test the actual functionality in
    # ConnectionManagerData.set_connection_options()
    with patch.object(pygw.connection, "set_connection_config") as set_config:
        pygw.set_connection_options(just="some", random="kwargs")
    set_config.assert_called_once_with(just="some", random="kwargs")


@pytest.mark.asyncio
async def test_set_target_temp(pygw):
    """Test pyotgw.set_target_temp()"""
    with pytest.raises(TypeError):
        await pygw.set_target_temp(None)

    with patch.object(pygw, "_wait_for_cmd", return_value=None) as wait_for_cmd:
        assert await pygw.set_target_temp(12.3) is None

    wait_for_cmd.assert_called_once_with(
        v.OTGW_CMD_TARGET_TEMP,
        "12.3",
        v.OTGW_DEFAULT_TIMEOUT,
    )

    with patch.object(
        pygw,
        "_wait_for_cmd",
        return_value="0.00",
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_full_update",
    ) as update_full_status:
        temp = await pygw.set_target_temp(0, timeout=5)

    assert isinstance(temp, float)
    assert temp == 0
    wait_for_cmd.assert_called_once_with(
        v.OTGW_CMD_TARGET_TEMP,
        "0.0",
        5,
    )
    update_full_status.assert_called_once_with(
        {
            v.OTGW: {v.OTGW_SETP_OVRD_MODE: v.OTGW_SETP_OVRD_DISABLED},
            v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: None},
        }
    )

    with patch.object(
        pygw,
        "_wait_for_cmd",
        return_value="15.50",
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_full_update",
    ) as update_full_status:
        temp = await pygw.set_target_temp(15.5)

    assert temp == 15.5
    wait_for_cmd.assert_called_once_with(v.OTGW_CMD_TARGET_TEMP, "15.5", 3)
    update_full_status.assert_called_once_with(
        {
            v.OTGW: {v.OTGW_SETP_OVRD_MODE: v.OTGW_SETP_OVRD_TEMPORARY},
            v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: 15.5},
        }
    )

    with patch.object(
        pygw,
        "_wait_for_cmd",
        return_value="20.50",
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_full_update",
    ) as update_full_status:
        temp = await pygw.set_target_temp(20.5, temporary=False)

    assert temp == 20.5
    wait_for_cmd.assert_called_once_with(v.OTGW_CMD_TARGET_TEMP_CONST, "20.5", 3)
    update_full_status.assert_called_once_with(
        {
            v.OTGW: {v.OTGW_SETP_OVRD_MODE: v.OTGW_SETP_OVRD_PERMANENT},
            v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: 20.5},
        }
    )


@pytest.mark.asyncio
async def test_set_temp_sensor_function(pygw):
    """Test pyotgw.set_temp_sensor_function()"""
    assert await pygw.set_temp_sensor_function("P") is None

    with patch.object(pygw, "_wait_for_cmd", return_value=None) as wait_for_cmd:
        assert await pygw.set_temp_sensor_function("O") is None

    wait_for_cmd.assert_called_once_with(
        v.OTGW_CMD_TEMP_SENSOR,
        "O",
        v.OTGW_DEFAULT_TIMEOUT,
    )

    with patch.object(
        pygw,
        "_wait_for_cmd",
        return_value="R",
    ) as wait_for_cmd, patch.object(
        pygw.status, "submit_partial_update"
    ) as update_status:
        assert await pygw.set_temp_sensor_function("R", timeout=5) == "R"

    wait_for_cmd.assert_called_once_with(v.OTGW_CMD_TEMP_SENSOR, "R", 5)
    update_status.assert_called_once_with(v.OTGW, {v.OTGW_TEMP_SENSOR: "R"})


@pytest.mark.asyncio
async def test_set_outside_temp(pygw):
    """Test pyotgw.set_outside_temp()"""
    assert await pygw.set_outside_temp(-40.1) is None

    with pytest.raises(TypeError):
        await pygw.set_outside_temp(None)

    with patch.object(pygw, "_wait_for_cmd", return_value=None) as wait_for_cmd:
        assert await pygw.set_outside_temp(0, timeout=5) is None

    wait_for_cmd.assert_called_once_with(v.OTGW_CMD_OUTSIDE_TEMP, "0.0", 5)

    with patch.object(
        pygw,
        "_wait_for_cmd",
        return_value="23.5",
    ) as wait_for_cmd, patch.object(
        pygw.status, "submit_partial_update"
    ) as update_status:
        assert await pygw.set_outside_temp(23.5) == 23.5

    wait_for_cmd.assert_called_once_with(
        v.OTGW_CMD_OUTSIDE_TEMP, "23.5", v.OTGW_DEFAULT_TIMEOUT
    )
    update_status.assert_called_once_with(v.THERMOSTAT, {v.DATA_OUTSIDE_TEMP: 23.5})

    with patch.object(
        pygw,
        "_wait_for_cmd",
        return_value="-",
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status:
        assert await pygw.set_outside_temp(99) == "-"

    wait_for_cmd.assert_called_once_with(
        v.OTGW_CMD_OUTSIDE_TEMP, "99.0", v.OTGW_DEFAULT_TIMEOUT
    )
    update_status.assert_called_once_with(v.THERMOSTAT, {v.DATA_OUTSIDE_TEMP: 0.0})


@pytest.mark.asyncio
async def test_set_clock(pygw):
    """Test pyotgw.set_clock()"""
    dt = datetime(year=2021, month=3, day=12, hour=12, minute=34)

    with patch.object(pygw, "_wait_for_cmd", return_value="12:34/5") as wait_for_cmd:
        assert await pygw.set_clock(dt) == "12:34/5"

    wait_for_cmd.assert_called_once_with(
        v.OTGW_CMD_SET_CLOCK, "12:34/5", v.OTGW_DEFAULT_TIMEOUT
    )

    with patch.object(pygw, "_wait_for_cmd", return_value="12:34/5") as wait_for_cmd:
        assert await pygw.set_clock(dt, timeout=5) == "12:34/5"

    wait_for_cmd.assert_called_once_with(v.OTGW_CMD_SET_CLOCK, "12:34/5", 5)


@pytest.mark.asyncio
async def test_get_reports(pygw):
    """Test pyotgw.get_reports()"""

    def get_response_42(cmd, val):
        """Get response from dict or raise ValueError"""
        try:
            return pygw_reports.report_responses_42[val]
        except KeyError:
            raise ValueError

    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=lambda _, v: pygw_reports.report_responses_51[v],
    ):
        assert await pygw.get_reports() == pygw_reports.expect_51

    pygw.status.reset()
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=get_response_42,
    ):
        assert await pygw.get_reports() == pygw_reports.expect_42

    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=["OpenTherm Gateway 5.1", ValueError],
    ), pytest.raises(ValueError):
        await pygw.get_reports()


@pytest.mark.asyncio
async def test_get_status(pygw):
    """Test pyotgw.get_status()"""
    pygw.loop = asyncio.get_running_loop()
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, (None, pygw_status.status_5), (None, pygw_status.status_4)],
    ):
        assert await pygw.get_status() is None
        assert await pygw.get_status() == pygw_status.expect_5
        pygw.status.reset()
        assert await pygw.get_status() == pygw_status.expect_4


@pytest.mark.asyncio
async def test_set_hot_water_ovrd(pygw):
    """Test pyotgw.set_hot_water_ovrd()"""
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, "A", "1"],
    ) as wait_for_cmd:
        assert await pygw.set_hot_water_ovrd(0) is None
        assert await pygw.set_hot_water_ovrd("A", 5) == "A"
        assert await pygw.set_hot_water_ovrd(1) == 1

    assert wait_for_cmd.call_count == 3
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_HOT_WATER, 0, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_HOT_WATER, "A", 5),
            call(v.OTGW_CMD_HOT_WATER, 1, v.OTGW_DEFAULT_TIMEOUT),
        ],
        any_order=False,
    )


@pytest.mark.asyncio
async def test_set_mode(pygw):
    """Test pyotgw.set_mode()"""
    with patch.object(pygw, "_wait_for_cmd", side_effect=[None, v.OTGW_MODE_MONITOR]):
        assert await pygw.set_mode(v.OTGW_MODE_GATEWAY) is None
        assert await pygw.set_mode(v.OTGW_MODE_MONITOR) == v.OTGW_MODE_MONITOR

    with patch.object(
        pygw,
        "_wait_for_cmd",
        return_value=v.OTGW_MODE_RESET,
    ), patch.object(pygw, "get_reports") as get_reports, patch.object(
        pygw,
        "get_status",
    ) as get_status:
        assert await pygw.set_mode(v.OTGW_MODE_RESET) == v.DEFAULT_STATUS
        get_reports.assert_called_once()
        get_status.assert_called_once()


@pytest.mark.asyncio
async def test_set_led_mode(pygw):
    """Test pyotgw.set_led_mode()"""
    assert await pygw.set_led_mode("G", "A") is None

    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, "X"],
    ) as wait_for_cmd, patch.object(
        pygw.status, "submit_partial_update"
    ) as update_status:
        assert await pygw.set_led_mode("B", "H") is None
        assert await pygw.set_led_mode("A", "X", timeout=5) == "X"

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_LED_B, "H", v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_LED_A, "X", 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.OTGW, {v.OTGW_LED_A: "X"})


@pytest.mark.asyncio
async def test_set_gpio_mode(pygw):
    """Test pyotgw.set_gpio_mode()"""
    assert await pygw.set_gpio_mode("A", 9) is None

    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 3],
    ) as wait_for_cmd, patch.object(
        pygw.status, "submit_partial_update"
    ) as update_status:
        assert await pygw.set_gpio_mode("A", 7) is None
        assert await pygw.set_gpio_mode("A", 6) is None
        assert await pygw.set_gpio_mode("B", 3, timeout=5) == 3

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_GPIO_A, 6, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_GPIO_B, 3, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.OTGW, {v.OTGW_GPIO_B: 3})


@pytest.mark.asyncio
async def test_set_setback_temp(pygw):
    """Test pyotgw.set_setback_temp()"""
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 16.5],
    ) as wait_for_cmd, patch.object(
        pygw.status, "submit_partial_update"
    ) as update_status:
        assert await pygw.set_setback_temp(17.5) is None
        assert await pygw.set_setback_temp(16.5, 5) == 16.5

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_SETBACK, 17.5, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_SETBACK, 16.5, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.OTGW, {v.OTGW_SB_TEMP: 16.5})


@pytest.mark.asyncio
async def test_add_alternative(pygw):
    """Test pyotgw.add_alternative()"""
    assert await pygw.add_alternative(0) is None

    with patch.object(pygw, "_wait_for_cmd", side_effect=[None, 23]) as wait_for_cmd:
        assert await pygw.add_alternative(20) is None
        assert await pygw.add_alternative(23, 5) == 23

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_ADD_ALT, 20, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_ADD_ALT, 23, 5),
        ],
        any_order=False,
    )


@pytest.mark.asyncio
async def test_del_alternative(pygw):
    """Test pyotgw.del_alternative()"""
    assert await pygw.del_alternative(0) is None

    with patch.object(pygw, "_wait_for_cmd", side_effect=[None, 23]) as wait_for_cmd:
        assert await pygw.del_alternative(20) is None
        assert await pygw.del_alternative(23, 5) == 23

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_DEL_ALT, 20, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_DEL_ALT, 23, 5),
        ],
        any_order=False,
    )


@pytest.mark.asyncio
async def test_add_unknown_id(pygw):
    """Test pyotgw.add_unknown_id()"""
    assert await pygw.add_unknown_id(0) is None

    with patch.object(pygw, "_wait_for_cmd", side_effect=[None, 23]) as wait_for_cmd:
        assert await pygw.add_unknown_id(20) is None
        assert await pygw.add_unknown_id(23, 5) == 23

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_UNKNOWN_ID, 20, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_UNKNOWN_ID, 23, 5),
        ],
        any_order=False,
    )


@pytest.mark.asyncio
async def test_del_unknown_id(pygw):
    """Test pyotgw.del_unknown_id()"""
    assert await pygw.del_unknown_id(0) is None

    with patch.object(pygw, "_wait_for_cmd", side_effect=[None, 23]) as wait_for_cmd:
        assert await pygw.del_unknown_id(20) is None
        assert await pygw.del_unknown_id(23, 5) == 23

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_KNOWN_ID, 20, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_KNOWN_ID, 23, 5),
        ],
        any_order=False,
    )


@pytest.mark.asyncio
async def test_set_max_ch_setpoint(pygw):
    """Test pyotgw.set_max_ch_setpoint()"""
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 74.5],
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status:
        assert await pygw.set_max_ch_setpoint(75.5) is None
        assert await pygw.set_max_ch_setpoint(74.5, 5) == 74.5

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_SET_MAX, 75.5, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_SET_MAX, 74.5, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.BOILER, {v.DATA_MAX_CH_SETPOINT: 74.5})


@pytest.mark.asyncio
async def test_set_dhw_setpoint(pygw):
    """Test pyotgw.set_dhw_setpoint()"""
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 54.5],
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status:
        assert await pygw.set_dhw_setpoint(55.5) is None
        assert await pygw.set_dhw_setpoint(54.5, 5) == 54.5

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_SET_WATER, 55.5, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_SET_WATER, 54.5, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.BOILER, {v.DATA_DHW_SETPOINT: 54.5})


@pytest.mark.asyncio
async def test_set_max_relative_mod(pygw):
    """Test pyotgw.set_max_relative_mod()"""
    assert await pygw.set_max_relative_mod(-1) is None

    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, "-", 55],
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status:
        assert await pygw.set_max_relative_mod(56) is None
        assert await pygw.set_max_relative_mod(54, 5) == "-"
        assert await pygw.set_max_relative_mod(55) == 55

    assert wait_for_cmd.call_count == 3
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_MAX_MOD, 56, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_MAX_MOD, 54, 5),
            call(v.OTGW_CMD_MAX_MOD, 55, v.OTGW_DEFAULT_TIMEOUT),
        ],
        any_order=False,
    )

    assert update_status.call_count == 2
    update_status.assert_has_calls(
        [
            call(v.BOILER, {v.DATA_SLAVE_MAX_RELATIVE_MOD: None}),
            call(v.BOILER, {v.DATA_SLAVE_MAX_RELATIVE_MOD: 55}),
        ],
        any_order=False,
    )


@pytest.mark.asyncio
async def test_set_control_setpoint(pygw):
    """Test pyotgw.set_control_setpoint()"""
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 19.5],
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status:
        assert await pygw.set_control_setpoint(21.5) is None
        assert await pygw.set_control_setpoint(19.5, 5) == 19.5

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_CONTROL_SETPOINT, 21.5, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_CONTROL_SETPOINT, 19.5, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.BOILER, {v.DATA_CONTROL_SETPOINT: 19.5})


@pytest.mark.asyncio
async def test_set_control_setpoint_2(pygw):
    """Test pyotgw.set_control_setpoint_2()"""
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 19.5],
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status:
        assert await pygw.set_control_setpoint_2(21.5) is None
        assert await pygw.set_control_setpoint_2(19.5, 5) == 19.5

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_CONTROL_SETPOINT_2, 21.5, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_CONTROL_SETPOINT_2, 19.5, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.BOILER, {v.DATA_CONTROL_SETPOINT_2: 19.5})


@pytest.mark.asyncio
async def test_set_ch_enable_bit(pygw):
    """Test pyotgw.set_ch_enable_bit()"""
    assert await pygw.set_ch_enable_bit(None) is None

    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 1],
    ) as wait_for_cmd, patch.object(
        pygw.status, "submit_partial_update"
    ) as update_status:
        assert await pygw.set_ch_enable_bit(0) is None
        assert await pygw.set_ch_enable_bit(1, 5) == 1

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_CONTROL_HEATING, 0, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_CONTROL_HEATING, 1, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.BOILER, {v.DATA_MASTER_CH_ENABLED: 1})


@pytest.mark.asyncio
async def test_set_ch2_enable_bit(pygw):
    """Test pyotgw.set_ch2_enable_bit()"""
    assert await pygw.set_ch2_enable_bit(None) is None

    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 1],
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status:
        assert await pygw.set_ch2_enable_bit(0) is None
        assert await pygw.set_ch2_enable_bit(1, 5) == 1

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_CONTROL_HEATING_2, 0, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_CONTROL_HEATING_2, 1, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.BOILER, {v.DATA_MASTER_CH2_ENABLED: 1})


@pytest.mark.asyncio
async def test_set_ventilation(pygw):
    """Test pyotgw.set_ventilation()"""
    assert await pygw.set_ventilation(-1) is None

    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=[None, 75],
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status:
        assert await pygw.set_ventilation(25) is None
        assert await pygw.set_ventilation(75, 5) == 75

    assert wait_for_cmd.call_count == 2
    wait_for_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_VENT, 25, v.OTGW_DEFAULT_TIMEOUT),
            call(v.OTGW_CMD_VENT, 75, 5),
        ],
        any_order=False,
    )

    update_status.assert_called_once_with(v.BOILER, {v.DATA_COOLING_CONTROL: 75})


@pytest.mark.asyncio
async def test_send_transparent_command(pygw):
    """Test pyotgw.send_transparent_command()"""
    with patch.object(
        pygw,
        "_wait_for_cmd",
        side_effect=["CD"],
    ) as wait_for_cmd:
        assert await pygw.send_transparent_command("AB", "CD") == "CD"

    assert wait_for_cmd.call_count == 1
    wait_for_cmd.assert_has_awaits(
        [
            call("AB", "CD", v.OTGW_DEFAULT_TIMEOUT),
        ],
    )


def test_subscribe_and_unsubscribe(pygw):
    """Test pyotgw.subscribe() and pyotgw.unsubscribe()"""

    async def empty_coroutine(status):
        return

    async def empty_coroutine_2(status):
        return

    assert pygw.subscribe(empty_coroutine)
    assert pygw.subscribe(empty_coroutine_2)
    assert not pygw.subscribe(empty_coroutine_2)

    assert pygw.status._notify == [empty_coroutine, empty_coroutine_2]

    assert pygw.unsubscribe(empty_coroutine)
    assert not pygw.unsubscribe(empty_coroutine)

    assert pygw.status._notify == [empty_coroutine_2]


@pytest.mark.asyncio
async def test_wait_for_cmd(caplog, pygw, pygw_proto):
    """Test pyotgw.wait_for_cmd()"""
    assert await pygw._wait_for_cmd(None, None) is None

    with patch(
        "pyotgw.connection.ConnectionManager.connected",
        return_value=True,
    ), patch.object(
        pygw_proto.command_processor,
        "issue_cmd",
        side_effect=[None, "0", asyncio.TimeoutError, ValueError],
    ) as issue_cmd, caplog.at_level(
        logging.ERROR
    ):
        assert await pygw._wait_for_cmd(v.OTGW_CMD_MODE, "G") is None
        assert await pygw._wait_for_cmd(v.OTGW_CMD_SUMMARY, 0) == "0"
        assert await pygw._wait_for_cmd(v.OTGW_CMD_REPORT, "I", 1) is None
        assert await pygw._wait_for_cmd(v.OTGW_CMD_MAX_MOD, -1) is None

    assert issue_cmd.await_count == 4
    issue_cmd.assert_has_awaits(
        [
            call(v.OTGW_CMD_MODE, "G"),
            call(v.OTGW_CMD_SUMMARY, 0),
            call(v.OTGW_CMD_REPORT, "I"),
            call(v.OTGW_CMD_MAX_MOD, -1),
        ],
        any_order=False,
    )

    assert caplog.record_tuples == [
        (
            "pyotgw.pyotgw",
            logging.ERROR,
            f"Timed out waiting for command: {v.OTGW_CMD_REPORT}, value: I.",
        ),
        (
            "pyotgw.pyotgw",
            logging.ERROR,
            f"Command {v.OTGW_CMD_MAX_MOD} with value -1 raised exception: ",
        ),
    ]


@pytest.mark.asyncio
async def test_poll_gpio(caplog, pygw):
    """Test pyotgw._poll_gpio()"""
    pygw._gpio_task = None
    pygw.loop = asyncio.get_running_loop()
    pygw.status.submit_partial_update(v.OTGW, {v.OTGW_GPIO_A: 4, v.OTGW_GPIO_B: 1})

    with caplog.at_level(logging.DEBUG):
        await pygw._poll_gpio()
    assert len(caplog.records) == 0

    pygw.status.submit_partial_update(v.OTGW, {v.OTGW_GPIO_B: 0})
    with patch.object(
        pygw,
        "_wait_for_cmd",
        return_value="I=10",
    ) as wait_for_cmd, patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status, caplog.at_level(
        logging.DEBUG
    ):
        await pygw._poll_gpio()
        await called_once(update_status)

    assert isinstance(pygw._gpio_task, asyncio.Task)
    wait_for_cmd.assert_awaited_once_with(v.OTGW_CMD_REPORT, v.OTGW_REPORT_GPIO_STATES)
    update_status.assert_called_once_with(
        v.OTGW,
        {v.OTGW_GPIO_A_STATE: 1, v.OTGW_GPIO_B_STATE: 0},
    )
    assert caplog.record_tuples == [
        ("pyotgw.pyotgw", logging.DEBUG, "Starting GPIO polling routine"),
    ]

    caplog.clear()
    pygw.status.submit_partial_update(v.OTGW, {v.OTGW_GPIO_B: 1})
    with patch.object(
        pygw.status,
        "submit_partial_update",
    ) as update_status, caplog.at_level(logging.DEBUG):
        await pygw._poll_gpio()
        await called_once(update_status)

    assert pygw._gpio_task is None
    update_status.assert_called_once_with(
        v.OTGW,
        {v.OTGW_GPIO_A_STATE: 0, v.OTGW_GPIO_B_STATE: 0},
    )
    assert caplog.record_tuples == [
        ("pyotgw.pyotgw", logging.DEBUG, "Stopping GPIO polling routine"),
        ("pyotgw.pyotgw", logging.DEBUG, "GPIO polling routine stopped"),
    ]



"""
Connection Manager for pyotgw.
Everything related to making, maintaining and monitoring the connection
to the gateway goes here.
"""

import asyncio
import logging
from functools import partial

import serial
import serial_asyncio_fast

from pyotgw.protocol import OpenThermProtocol

CONNECTION_TIMEOUT = 5

MAX_RETRY_TIMEOUT = 60
MIN_RETRY_TIMEOUT = 5

WATCHDOG_TIMEOUT = 3

_LOGGER = logging.getLogger(__name__)


class ConnectionManager:  # pylint: disable=too-many-instance-attributes
    """Functionality for setting up and tearing down a connection"""

    def __init__(self, otgw):
        """Initialise the connection manager"""
        self._error = None
        self._port = None
        self._retry_timeout = MIN_RETRY_TIMEOUT
        self._connecting_task = None
        self._config = {
            "baudrate": 9600,
            "bytesize": serial.EIGHTBITS,
            "parity": serial.PARITY_NONE,
            "stopbits": serial.STOPBITS_ONE,
        }
        self._otgw = otgw
        self._timeout = 10
        self.watchdog = ConnectionWatchdog()
        self._transport = None
        self.protocol = None

    async def connect(self, port, timeout=None):
        """Start connection attempts. Return True on success or False on failure."""
        if self.connected or self._connecting_task:
            # We are actually reconnecting, cleanup first.
            _LOGGER.debug("Reconnecting to serial device on %s", port)
            await self.disconnect()

        loop = asyncio.get_running_loop()
        self._port = port
        self._timeout = timeout if timeout else self._timeout
        self._connecting_task = loop.create_task(self._attempt_connect())
        try:
            transport, protocol = await self._connecting_task
        except asyncio.CancelledError:
            return False
        finally:
            self._connecting_task = None
        self._error = None
        _LOGGER.debug("Connected to serial device on %s", port)
        self._transport = transport
        self.protocol = protocol
        self.watchdog.start(self.reconnect, timeout=timeout or WATCHDOG_TIMEOUT)
        return True

    async def disconnect(self):
        """Disconnect from the OpenTherm Gateway."""
        await self._cleanup()
        if self.connected:
            self.protocol.disconnect()

    async def reconnect(self):
        """Reconnect to the OpenTherm Gateway."""
        if not self._port:
            _LOGGER.error("Reconnect called before connect!")
            return
        _LOGGER.debug("Scheduling reconnect...")
        await self.disconnect()
        await self._otgw.connect(self._port)

    @property
    def connected(self):
        """Return the connection status"""
        return self.protocol and self.protocol.connected

    def set_connection_config(self, **kwargs):
        """
        Set the serial connection parameters before calling connect()
        Valid kwargs are 'baudrate', 'bytesize', 'parity' and 'stopbits'.
        Returns True on success, False on fail or if already connected.
        For more information see the pyserial documentation.
        """
        if self.connected:
            return False
        for arg in kwargs:
            if arg not in self._config:
                _LOGGER.error("Invalid connection parameter: %s", arg)
                return False
        self._config.update(kwargs)
        return True

    async def _attempt_connect(self):
        """Try to connect to the OpenTherm Gateway."""
        loop = asyncio.get_running_loop()
        transport = None
        protocol = None
        self._retry_timeout = MIN_RETRY_TIMEOUT
        while transport is None:
            try:
                (
                    transport,
                    protocol,
                ) = await serial_asyncio_fast.create_serial_connection(
                    loop,
                    partial(
                        OpenThermProtocol,
                        self._otgw.status,
                        self.watchdog.inform,
                    ),
                    self._port,
                    write_timeout=0,
                    **self._config,
                )
                await asyncio.wait_for(
                    protocol.init_and_wait_for_activity(),
                    CONNECTION_TIMEOUT,
                )
                return transport, protocol

            except asyncio.TimeoutError as err:
                if not isinstance(err, type(self._error)):
                    _LOGGER.error(
                        "The serial device on %s is not responding. "
                        "Will keep trying.",
                        self._port,
                    )
                    self._error = err
                if protocol:
                    await protocol.cleanup()

            except (serial.SerialException, OSError) as err:
                if not isinstance(err, type(self._error)):
                    _LOGGER.error(
                        "Could not connect to serial device on %s. "
                        "Will keep trying. Reported error was: %s",
                        self._port,
                        err,
                    )
                    self._error = err
                if protocol:
                    await protocol.cleanup()

            except SyntaxError as err:
                # The gateway may throw a SyntaxError on the first initialization
                # attempt.
                if protocol:
                    await protocol.cleanup()
                if isinstance(err, type(self._error)):
                    raise
                self._error = err

            transport = None
            await asyncio.sleep(self._get_retry_timeout())

    async def _cleanup(self):
        """Cleanup possible leftovers from old connections"""
        await self.watchdog.stop()
        self._otgw.status.reset()
        if self.protocol:
            await self.protocol.cleanup()
        if self._connecting_task is not None:
            self._connecting_task.cancel()
            try:
                await self._connecting_task
            except asyncio.CancelledError:
                self._connecting_task = None

    def _get_retry_timeout(self):
        """Increase if needed and return the retry timeout."""
        if self._retry_timeout == MAX_RETRY_TIMEOUT:
            return self._retry_timeout
        timeout = self._retry_timeout
        self._retry_timeout = min([self._retry_timeout * 1.5, MAX_RETRY_TIMEOUT])
        return timeout


class ConnectionWatchdog:
    """Connection watchdog"""

    def __init__(self):
        """Initialise the object"""
        self._callback = None
        self.timeout = WATCHDOG_TIMEOUT
        self._wd_task = None
        self._lock = asyncio.Lock()
        self.loop = asyncio.get_event_loop()

    @property
    def is_active(self):
        """Return watchdog status"""
        return self._wd_task is not None

    async def inform(self):
        """Reset the watchdog timer."""
        async with self._lock:
            if not self.is_active:
                # Check within the Lock to deal with external stop()
                # calls with queued inform() tasks.
                return
            self._wd_task.cancel()
            try:
                await self._wd_task
            except asyncio.CancelledError:
                self._wd_task = self.loop.create_task(self._watchdog(self.timeout))
                _LOGGER.debug("Watchdog timer reset!")

    def start(self, callback, timeout):
        """Start the watchdog, return boolean indicating success"""
        if self.is_active:
            return False
        self._callback = callback
        self.timeout = timeout
        self._wd_task = self.loop.create_task(self._watchdog(timeout))
        return self.is_active

    async def stop(self):
        """Stop the watchdog"""
        async with self._lock:
            if not self.is_active:
                return
            _LOGGER.debug("Canceling Watchdog task.")
            self._wd_task.cancel()
            try:
                await self._wd_task
            except asyncio.CancelledError:
                self._wd_task = None

    async def _watchdog(self, timeout):
        """Trigger and cancel the watchdog after timeout. Schedule callback."""
        await asyncio.sleep(timeout)
        _LOGGER.debug("Watchdog triggered!")
        self.loop.create_task(self._callback())
        await self.stop()



"""Asyncio protocol implementation for pyotgw"""

import asyncio
import logging
import re

from pyotgw import vars as v
from pyotgw.commandprocessor import CommandProcessor
from pyotgw.messageprocessor import MessageProcessor

_LOGGER = logging.getLogger(__name__)


class OpenThermProtocol(
    asyncio.Protocol
):  # pylint: disable=too-many-instance-attributes
    """
    Implementation of the Opentherm Gateway protocol to be used with
    asyncio connections.
    """

    def __init__(
        self,
        status_manager,
        activity_callback,
    ):
        """Initialise the protocol object."""
        self.transport = None
        self._readbuf = b""
        self._received_lines = 0
        self.activity_callback = activity_callback
        self.command_processor = CommandProcessor(
            self,
            status_manager,
        )
        self._connected = False
        self.message_processor = MessageProcessor(
            self.command_processor,
            status_manager,
        )
        self.status_manager = status_manager

    def connection_made(self, transport):
        """Gets called when a connection to the gateway is established."""
        self.transport = transport
        self._received_lines = 0
        self._connected = True

    def connection_lost(self, exc):
        """
        Gets called when the connection to the gateway is lost.
        Tear down and clean up the protocol object.
        """
        if self.active and self.connected:
            _LOGGER.error("Disconnected: %s", exc)
        self._received_lines = 0
        self._connected = False
        self.command_processor.clear_queue()
        self.message_processor.connection_lost()
        self.status_manager.reset()

    @property
    def connected(self):
        """Return the connection status"""
        return self._connected

    async def cleanup(self):
        """Clean up"""
        self.disconnect()
        await self.message_processor.cleanup()

    def disconnect(self):
        """Disconnect gracefully."""
        if self.transport.is_closing() or not self.connected:
            return
        self._connected = False
        self.transport.close()

    def data_received(self, data):
        """
        Gets called when new data is received on the serial interface.
        Perform line buffering and call line_received() with complete
        lines.
        """
        # DIY line buffering...
        newline = b"\r\n"
        eot = b"\x04"
        self._readbuf += data
        while newline in self._readbuf:
            line, _, self._readbuf = self._readbuf.partition(newline)
            if line:
                if eot in line:
                    # Discard everything before EOT
                    _, _, line = line.partition(eot)
                try:
                    decoded = line.decode("ascii")
                except UnicodeDecodeError:
                    _LOGGER.debug("Invalid data received, ignoring...")
                    return
                self.line_received(decoded)

    def line_received(self, line):
        """
        Gets called by data_received() when a complete line is
        received.
        Inspect the received line and process or queue accordingly.
        """
        self._received_lines += 1
        _LOGGER.debug("Received line %d: %s", self._received_lines, line)
        if self.activity_callback:
            asyncio.create_task(self.activity_callback())
        pattern = r"^(T|B|R|A|E)([0-9A-F]{8})$"
        msg = re.match(pattern, line)
        if msg:
            self.message_processor.submit_matched_message(msg)
        elif re.match(r"^[0-9A-F]{1,8}$", line) and self._received_lines == 1:
            # Partial message on fresh connection. Ignore.
            self._received_lines = 0
            _LOGGER.debug("Ignoring line: %s", line)
        else:
            _LOGGER.debug(
                "Submitting line %d to CommandProcessor",
                self._received_lines,
            )
            self.command_processor.submit_response(line)

    @property
    def active(self):
        """Indicate that we have seen activity on the serial line."""
        return self._received_lines > 0

    async def init_and_wait_for_activity(self):
        """Wait for activity on the serial connection."""
        await self.command_processor.issue_cmd(v.OTGW_CMD_SUMMARY, 0, retry=1)
        while not self.active:
            await asyncio.sleep(0)



"""Helper functions for tests"""

import asyncio


async def called_x_times(mocked, x, timeout=10):
    """Wait for x or more calls on mocked object or timeout"""

    async def _wait():
        while mocked.call_count < x:
            await asyncio.sleep(0)

    await asyncio.wait_for(_wait(), timeout)


async def called_once(mocked, timeout=10):
    """Wait for at least 1 call on mocked object or timeout"""
    await called_x_times(mocked, 1, timeout)


async def let_queue_drain(queue, timeout=10):
    """Wait for queue to become empty or timeout"""

    async def _wait():
        while not queue.empty():
            await asyncio.sleep(0)

    await asyncio.wait_for(_wait(), timeout)



"""OpenTherm Gateway command handler."""

import asyncio
import logging
import re
from asyncio.queues import QueueFull

from pyotgw import vars as v

_LOGGER = logging.getLogger(__name__)


class CommandProcessor:
    """OpenTherm Gateway command handler."""

    def __init__(
        self,
        protocol,
        status_manager,
    ):
        """Initialise the CommandProcessor object."""
        self.protocol = protocol
        self._lock = asyncio.Lock()
        self._cmdq = asyncio.Queue()
        self.status_manager = status_manager

    async def issue_cmd(self, cmd, value, retry=3):
        """
        Issue a command, then await and return the return value.

        This method is a coroutine
        """
        async with self._lock:
            if not self.protocol.connected:
                _LOGGER.debug("Serial transport closed, not sending command %s", cmd)
                return
            self.clear_queue()
            if isinstance(value, float):
                value = f"{value:.2f}"
            _LOGGER.debug("Sending command: %s with value %s", cmd, value)
            self.protocol.transport.write(f"{cmd}={value}\r\n".encode("ascii"))
            expect = self._get_expected_response(cmd, value)

            async def send_again(err):
                """Resend the command."""
                nonlocal retry
                _LOGGER.warning("Command %s failed with %s, retrying...", cmd, err)
                retry -= 1
                self.protocol.transport.write(f"{cmd}={value}\r\n".encode("ascii"))

            async def process(msg):
                """Process a possible response."""
                _LOGGER.debug("Got possible response for command %s: %s", cmd, msg)
                if msg in v.OTGW_ERRS:
                    # Some errors appear by themselves on one line.
                    if retry == 0:
                        raise v.OTGW_ERRS[msg]
                    await send_again(msg)
                    return
                if cmd == v.OTGW_CMD_MODE and value == "R":
                    # Device was reset, msg contains build info
                    while not re.match(r"OpenTherm Gateway \d+(\.\d+)*", msg):
                        msg = await self._cmdq.get()
                    return True
                match = re.match(expect, msg)
                if match:
                    if match.group(1) in v.OTGW_ERRS:
                        # Some errors are considered a response.
                        if retry == 0:
                            raise v.OTGW_ERRS[match.group(1)]
                        await send_again(msg)
                        return
                    ret = match.group(1)
                    if cmd == v.OTGW_CMD_SUMMARY and ret == "1":
                        # Expects a second line
                        part2 = await self._cmdq.get()
                        ret = [ret, part2]
                    return ret
                if re.match(r"Error 0[1-4]", msg):
                    _LOGGER.warning(
                        "Received %s. If this happens during a "
                        "reset of the gateway it can be safely "
                        "ignored.",
                        msg,
                    )
                else:
                    _LOGGER.warning("Unknown message in command queue: %s", msg)
                await send_again(msg)

            while True:
                msg = await self._cmdq.get()
                ret = await process(msg)
                if ret is not None:
                    return ret

    def clear_queue(self):
        """Clear leftover messages from the command queue"""
        while not self._cmdq.empty():
            _LOGGER.debug(
                "Clearing leftover message from command queue: %s",
                self._cmdq.get_nowait(),
            )

    def submit_response(self, response):
        """Add a possible response to the command queue"""
        try:
            self._cmdq.put_nowait(response)
            _LOGGER.debug("Response submitted. Queue size: %d", self._cmdq.qsize())
        except QueueFull:
            _LOGGER.error("Queue full, discarded message: %s", response)

    @staticmethod
    def _get_expected_response(cmd, value):
        """Return the expected response pattern"""
        if cmd == v.OTGW_CMD_REPORT:
            return rf"^{cmd}:\s*([A-Z]{{2}}|{value}=[^$]+)$"
        # OTGW_CMD_CONTROL_HEATING_2 and OTGW_CMD_CONTROL_SETPOINT_2 do not adhere
        # to the standard response format (<cmd>: <value>) at the moment, but report
        # only the value. This will likely be fixed in the future, so we support
        # both formats.
        if cmd in (
            v.OTGW_CMD_CONTROL_HEATING_2,
            v.OTGW_CMD_CONTROL_SETPOINT_2,
        ):
            return rf"^(?:{cmd}:\s*)?(0|1|[0-9]+\.[0-9]{{2}}|[A-Z]{{2}})$"
        return rf"^{cmd}:\s*([^$]+)$"
